# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd
import itertools
from collections import defaultdict
from copy import deepcopy
from dataclasses import dataclass, field
from functools import reduce
from operator import mul
from typing import Dict, List, Optional, Set, Tuple, Union

import numpy as np
from compiler_config.config import InlineResultsProcessing

from qat.purr.backends.qblox.constants import Constants
from qat.purr.backends.qblox.graph import ControlFlowGraph
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.devices import PulseChannel, PulseShapeType
from qat.purr.compiler.instructions import (
    Acquire,
    Assign,
    Delay,
    DeviceUpdate,
    EndRepeat,
    EndSweep,
    Instruction,
    PostProcessing,
    PostProcessType,
    ProcessAxis,
    Pulse,
    QuantumInstruction,
    Repeat,
    ResultsProcessing,
    Return,
    Sweep,
    Variable,
    calculate_duration,
)
from qat.purr.core.pass_base import AnalysisPass
from qat.purr.core.result_base import ResultInfoMixin, ResultManager
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


@dataclass
class TriageResult(ResultInfoMixin):
    sweeps: List[Sweep] = field(default_factory=list)
    repeats: List[Repeat] = field(default_factory=list)
    returns: List[Return] = field(default_factory=list)
    assigns: List[Assign] = field(default_factory=list)
    target_map: Dict[PulseChannel, List[Instruction]] = field(
        default_factory=lambda: defaultdict(list)
    )
    device_updates: List[DeviceUpdate] = field(default_factory=list)
    quantum_instructions: List[QuantumInstruction] = field(default_factory=list)
    acquire_map: Dict[PulseChannel, List[Acquire]] = field(
        default_factory=lambda: defaultdict(list)
    )
    pp_map: Dict[str, List[PostProcessing]] = field(
        default_factory=lambda: defaultdict(list)
    )
    rp_map: Dict[str, ResultsProcessing] = field(default_factory=dict)


class TriagePass(AnalysisPass):
    """
    Builds a view of instructions per quantum target AOT.

    Builds selections of instructions useful for subsequent analysis/transform passes, for
    code generation, and post-playback steps.

    This is equivalent to the :class:`QatFile` and simplifies the duration timeline creation
    in the legacy code.
    """

    def run(self, ir: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs):
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        :param res_mgr: The result manager to save the analysis results.
        """

        targets = set()
        for inst in ir.instructions:
            if isinstance(inst, QuantumInstruction):
                if isinstance(inst, PostProcessing):
                    for qt in inst.quantum_targets:
                        if isinstance(qt, Acquire):
                            targets.update(qt.quantum_targets)
                        else:
                            targets.add(qt)
                else:
                    targets.update(inst.quantum_targets)

        result = TriageResult()
        for inst in ir.instructions:
            # Dissect by target
            if isinstance(inst, QuantumInstruction):
                result.quantum_instructions.append(inst)
                for qt in inst.quantum_targets:
                    if isinstance(qt, Acquire):
                        for aqt in qt.quantum_targets:
                            result.target_map[aqt].append(inst)
                    else:
                        result.target_map[qt].append(inst)
            elif isinstance(inst, Sweep | EndSweep | Repeat | EndRepeat):
                for t in targets:
                    result.target_map[t].append(inst)

            # Sweeps
            if isinstance(inst, Sweep):
                result.sweeps.append(inst)

            # Repeats
            elif isinstance(inst, Repeat):
                result.repeats.append(inst)

            # DeviceUpdates
            elif isinstance(inst, DeviceUpdate):
                result.device_updates.append(inst)

            # Returns
            elif isinstance(inst, Return):
                result.returns.append(inst)

            # Assigns
            elif isinstance(inst, Assign):
                result.assigns.append(inst)

            # Acquisition by target
            elif isinstance(inst, Acquire):
                for t in inst.quantum_targets:
                    result.acquire_map[t].append(inst)

            # Post-processing by output variable
            elif isinstance(inst, PostProcessing):
                result.pp_map[inst.output_variable].append(inst)

            # Results-processing by output variable
            elif isinstance(inst, ResultsProcessing):
                result.rp_map[inst.variable] = inst

        # Assume that raw acquisitions are experiment results.
        acquires = list(itertools.chain(*result.acquire_map.values()))
        missing_results = {
            acq.output_variable
            for acq in acquires
            if acq.output_variable not in result.rp_map
        }
        for missing_var in missing_results:
            result.rp_map[missing_var] = ResultsProcessing(
                missing_var, InlineResultsProcessing.Experiment
            )

        res_mgr.add(result)


@dataclass
class ScopingResult:
    scope2symbols: Dict[Tuple[Instruction, Optional[Instruction]], Set[str]] = field(
        default_factory=lambda: defaultdict(set)
    )
    symbol2scopes: Dict[str, List[Tuple[Instruction, Optional[Instruction]]]] = field(
        default_factory=lambda: defaultdict(list)
    )


@dataclass
class ReadWriteResult:
    inits: Dict[str, List[Instruction]] = field(default_factory=lambda: defaultdict(list))
    reads: Dict[str, List[Instruction]] = field(default_factory=lambda: defaultdict(list))
    writes: Dict[str, List[Instruction]] = field(default_factory=lambda: defaultdict(list))


@dataclass(frozen=True)
class IterBound:
    start: Union[int, float, complex] = None
    step: Union[int, float, complex] = None
    end: Union[int, float, complex] = None
    count: int = None

    def astype(self, ty: type):
        return IterBound(
            start=np.array([self.start], dtype=type(self.start)).view(ty)[0],
            step=np.array([self.step], dtype=type(self.step)).view(ty)[0],
            end=np.array([self.end], dtype=type(self.end)).view(ty)[0],
            count=self.count,
        )


@dataclass
class BindingResult(ResultInfoMixin):
    scoping_results: Dict[PulseChannel, ScopingResult] = field(
        default_factory=lambda: defaultdict(lambda: ScopingResult())
    )
    rw_results: Dict[PulseChannel, ReadWriteResult] = field(
        default_factory=lambda: defaultdict(lambda: ReadWriteResult())
    )
    iter_bound_results: Dict[PulseChannel, Dict[str, IterBound]] = field(
        default_factory=lambda: defaultdict(dict)
    )


class BindingPass(AnalysisPass):
    """
    Builds binding of variables, instructions, and a view of variables from/to scopes.

    Variables are implicitly declared in sweep instructions and are ultimately read from
    quantum instructions. Thus, every iteration variable is associated to all the scopes it
    is declared in.

    Values of the iteration variable are abstract and don't mean anything. In this pass, we
    only extract  the bound and associate it to the name variable. Further analysis is
    required on read sites to make sure their usage is consistent and meaningful.
    """

    @staticmethod
    def extract_iter_bound(value: Union[List, np.ndarray]):
        """
        Given a sequence of numbers (typically having been generated from
        :code:`np.linspace()`), figure out if the numbers are linearly/evenly spaced,
        in which case returns an IterBound instance holding the start, step, end, and count
        of the numbers in the array, or else fail.

        In the future, we might be interested in relaxing this condition and return
        "interpolated" evenly spaced approximation of the input sequence.
        """

        if value is None:
            raise ValueError(f"Cannot process value {value}")

        if isinstance(value, np.ndarray):
            value = value.tolist()

        if not value:
            raise ValueError(f"Cannot process value {value}")

        start = value[0]
        step = 0
        end = value[-1]
        count = len(value)

        if count < 2:
            return IterBound(start, step, end, count)

        step = value[1] - value[0]

        if not np.isclose(step, (end - start) / (count - 1)):
            raise ValueError(f"Not a regularly partitioned space {value}")

        return IterBound(start, step, end, count)

    def run(self, ir: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs):
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        :param res_mgr: The result manager to save the analysis results.
        """

        triage_result = res_mgr.lookup_by_type(TriageResult)
        result = BindingResult()
        for target, instructions in triage_result.target_map.items():
            scoping_result = result.scoping_results[target]
            rw_result = result.rw_results[target]
            iter_bound_result = result.iter_bound_results[target]

            stack = []
            for inst in instructions:
                parent_scopes = [
                    (s, e) for (s, e) in scoping_result.scope2symbols if s in stack
                ]
                if isinstance(inst, Sweep):
                    stack.append(inst)
                    scope = (inst, None)
                    if p := next(iter(parent_scopes[::-1]), None):
                        for name in scoping_result.scope2symbols[p]:
                            if name not in scoping_result.scope2symbols[scope]:
                                scoping_result.scope2symbols[scope].add(name)
                                scoping_result.symbol2scopes[name].append(scope)

                    for name, value in inst.variables.items():
                        scoping_result.scope2symbols[scope].add(name)
                        scoping_result.symbol2scopes[name].append(scope)
                        iter_bound_result[name] = self.extract_iter_bound(value)

                    iter_name = f"{hash(inst)}"
                    rw_result.inits[iter_name].append(inst)
                    rw_result.reads[iter_name].append(inst)
                    rw_result.writes[iter_name].append(inst)
                elif isinstance(inst, Repeat):
                    stack.append(inst)
                    scope = (inst, None)
                    if p := next(iter(parent_scopes[::-1]), None):
                        for name in scoping_result.scope2symbols[p]:
                            if name not in scoping_result.scope2symbols[scope]:
                                scoping_result.scope2symbols[scope].add(name)
                                scoping_result.symbol2scopes[name].append(scope)

                    iter_name = f"{hash(inst)}"
                    scoping_result.scope2symbols[scope].add(iter_name)
                    scoping_result.symbol2scopes[iter_name].append(scope)
                    iter_bound_result[iter_name] = IterBound(
                        start=1, step=1, end=inst.repeat_count, count=inst.repeat_count
                    )

                    iter_name = f"{hash(inst)}"
                    rw_result.inits[iter_name].append(inst)
                    rw_result.reads[iter_name].append(inst)
                    rw_result.writes[iter_name].append(inst)
                elif isinstance(inst, (EndSweep, EndRepeat)):
                    try:
                        delimiter = stack.pop()
                    except IndexError:
                        raise ValueError(f"Unbalanced scope. Found orphan {inst}")

                    delimiter_type = Sweep if isinstance(inst, EndSweep) else Repeat
                    if not isinstance(delimiter, delimiter_type):
                        raise ValueError(f"Unbalanced scope. Found orphan {inst}")

                    scope = next(
                        (
                            (s, e)
                            for (s, e) in scoping_result.scope2symbols.keys()
                            if s == delimiter
                        )
                    )
                    symbols = scoping_result.scope2symbols[scope]
                    del scoping_result.scope2symbols[scope]
                    scoping_result.scope2symbols[(delimiter, inst)] = symbols

                    for name, scopes in scoping_result.symbol2scopes.items():
                        scoping_result.symbol2scopes[name] = [
                            (delimiter, inst) if s == scope else s for s in scopes
                        ]
                elif isinstance(inst, Acquire):
                    if any(
                        pp.process == PostProcessType.MEAN
                        and ProcessAxis.SEQUENCE in pp.axes
                        for pp in triage_result.pp_map[inst.output_variable]
                    ):
                        # Discard scopes defined by Repeat instructions
                        parent_scopes = [
                            (s, e) for (s, e) in parent_scopes if not isinstance(s, Repeat)
                        ]

                    iter_name = f"{hash(inst)}"
                    shape = tuple(
                        iter_bound_result[f"{hash(s)}"].count for (s, e) in parent_scopes
                    )
                    loop_nest_size = reduce(mul, shape, 1)
                    iter_bound_result[iter_name] = IterBound(
                        start=0, step=1, end=loop_nest_size, count=loop_nest_size
                    )

                    iter_name = f"{hash(inst)}"
                    rw_result.reads[iter_name].append(inst)
                    if (
                        innermost := next((s for (s, e) in parent_scopes[::-1]), None)
                    ) is not None:
                        rw_result.writes[iter_name].append(innermost)
                else:
                    attr2var = {
                        attr: var
                        for attr, var in vars(inst).items()
                        if isinstance(var, Variable)
                    }

                    for attr, var in attr2var.items():
                        defining_sweep = next(
                            (
                                s
                                for s in stack[::-1]
                                if isinstance(s, Sweep) and var.name in s.variables
                            ),
                            None,
                        )
                        if defining_sweep is None:
                            raise ValueError(
                                f"Variable {var} referenced but no prior declaration found in target {target}"
                            )

                        rw_result.inits[var.name].append(defining_sweep)
                        rw_result.reads[var.name].append(inst)
                        rw_result.writes[var.name].append(defining_sweep)

            if len(stack) > 0:
                raise ValueError(
                    f"Unbalanced scopes. Found orphans {stack} in target {target}"
                )

        res_mgr.add(result)


class TILegalisationPass(AnalysisPass):
    """
    An instruction is legal if it has a direct equivalent in the programming model
    implemented by the control stack. The notion of "legal" is highly determined by the
    hardware features of the control stack as well as its programming model. Control stacks
    such as Qblox have a direct ISA-level representation for basic RF instructions such as
    frequency and phase manipulation, arithmetic instructions such as add,  and branching
    instructions such as jump.

    This pass performs target-independent legalisation. The goal here is to understand how
    variables are used and legalise their bounds. Furthermore, analysis in this pass is
    fundamentally based on QAT semantics and must be kept target-agnostic so that it can be
    reused among backends.

    Particularly in QAT:
    #. A sweep instruction is illegal because it specifies unclear iteration semantics.
    #. Device updates/assigns in general are illegal because they are bound to a sweep
       instruction via a variable. In fact, a variable (implicitly defined by a Sweep
       instruction) remains obscure until a "read" (usually on the instruction builder or on
       the hardware model) (typically from a DeviceUpdate instruction) is encountered where
       its intent becomes clear. We say that a DeviceUpdate carries meaning for the variable
       and materialises its intention.
    """

    @staticmethod
    def decompose_freq(frequency: float, target: PulseChannel):
        if target.fixed_if:  # NCO freq constant
            nco_freq = target.baseband_if_frequency
            lo_freq = frequency - nco_freq
        else:  # LO freq constant
            lo_freq = target.baseband_frequency
            nco_freq = frequency - lo_freq

        return lo_freq, nco_freq

    @staticmethod
    def transform_amp(amp: float, scale_factor: float, ignore_scale, target: PulseChannel):
        bias = target.bias
        scale = (1.0 if ignore_scale else target.scale) + 0.0j
        pulse_amp = scale * (scale_factor * amp) + bias

        if abs(pulse_amp.real) > 1 or abs(pulse_amp.imag) > 1:
            raise ValueError("Illegal DAC/ADC ratio. It must be within range [-1, 1]")

        if pulse_amp.imag != 0:
            raise NotImplementedError("Unsupported processing of complex amplitudes")

        return pulse_amp.real

    def _legalise_bound(self, name: str, bound: IterBound, inst: Instruction):
        legal_bound = bound

        if isinstance(inst, Delay):
            legal_bound = IterBound(
                start=int(calculate_duration(Delay(inst.quantum_targets, bound.start))),
                step=int(calculate_duration(Delay(inst.quantum_targets, bound.step))),
                end=int(calculate_duration(Delay(inst.quantum_targets, bound.end))),
                count=bound.count,
            )
        elif isinstance(inst, DeviceUpdate):
            if inst.attribute not in ["frequency", "scale", "amp"]:
                raise NotImplementedError(
                    f"Unsupported processing of attribute {inst.attribute} for instruction {inst}"
                )

            if inst.attribute == "frequency":
                if inst.target.fixed_if:
                    raise ValueError(
                        f"fixed_if must be False on target {inst.target} to sweep over frequencies"
                    )
                legal_bound = IterBound(
                    start=self.decompose_freq(bound.start, inst.target)[1],
                    step=bound.step,
                    end=self.decompose_freq(bound.end, inst.target)[1],
                    count=bound.count,
                )
        else:
            attr2var = {
                attr: var
                for attr, var in vars(inst).items()
                if isinstance(var, Variable) and var.name == name
            }

            if not attr2var:
                return legal_bound
            if len(attr2var) > 1:
                raise ValueError(
                    f"Unsafe analysis. Distinct attributes expecting the same variable bound {attr2var}"
                )
            if isinstance(inst, Pulse) and inst.shape != PulseShapeType.SQUARE:
                raise ValueError("Cannot process non-trivial pulses")
            attr, var = next(iter(attr2var.items()))
            if attr not in ["width", "amp", "phase"]:
                raise NotImplementedError(
                    f"Unsupported processing of attribute {attr} for instruction {inst}"
                )

            target: PulseChannel = next(iter(inst.quantum_targets))
            if attr == "width":
                legal_bound = IterBound(
                    start=int(calculate_duration(Delay(target, bound.start))),
                    step=int(calculate_duration(Delay(target, bound.step))),
                    end=int(calculate_duration(Delay(target, bound.end))),
                    count=bound.count,
                )
            elif attr == "amp" and isinstance(inst, Pulse):
                legal_bound = IterBound(
                    start=self.transform_amp(
                        bound.start,
                        inst.scale_factor,
                        inst.ignore_channel_scale,
                        target,
                    ),
                    step=self.transform_amp(
                        bound.step, inst.scale_factor, inst.ignore_channel_scale, target
                    ),
                    end=self.transform_amp(
                        bound.end, inst.scale_factor, inst.ignore_channel_scale, target
                    ),
                    count=bound.count,
                )

        return legal_bound

    def run(self, ir: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs):
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        :param res_mgr: The result manager to save the analysis results.
        """

        triage_result = res_mgr.lookup_by_type(TriageResult)
        binding_result = res_mgr.lookup_by_type(BindingResult)

        for target in triage_result.target_map:
            rw_result = binding_result.rw_results[target]
            bound_result = binding_result.iter_bound_results[target]
            legal_bound_result: Dict[str, IterBound] = deepcopy(bound_result)

            read_bounds: Dict[str, Set[IterBound]] = defaultdict(set)
            for name, instructions in rw_result.reads.items():
                for inst in instructions:
                    legal_bound = self._legalise_bound(name, bound_result[name], inst)
                    read_bounds[name].add(legal_bound)

            for name, bound_set in read_bounds.items():
                if len(bound_set) > 1:
                    raise ValueError(
                        f"Ambiguous bounds for variable {name} in target {target}"
                    )
                legal_bound_result[name] = next(iter(bound_set))

            # TODO: the proper way is to produce a new result and invalidate the old one
            binding_result.iter_bound_results[target] = legal_bound_result


class QbloxLegalisationPass(AnalysisPass):
    @staticmethod
    def phase_as_steps(phase_rad: float) -> int:
        """
        The instruction `set_ph_delta` expects the phase shift as a (potentially signed) integer operand.
        """

        phase_deg = np.rad2deg(phase_rad)
        phase_deg %= 360
        steps = int(round(phase_deg * Constants.NCO_PHASE_STEPS_PER_DEG))
        return steps

    @staticmethod
    def freq_as_steps(freq_hz: float) -> int:
        """
        The instruction `set_freq` expects the frequency as a (potentially signed) integer operand.
        """

        steps = int(round(freq_hz * Constants.NCO_FREQ_STEPS_PER_HZ))

        if (
            steps < -Constants.NCO_FREQ_LIMIT_STEPS
            or steps > Constants.NCO_FREQ_LIMIT_STEPS
        ):
            min_max_frequency_in_hz = (
                Constants.NCO_FREQ_LIMIT_STEPS / Constants.NCO_FREQ_STEPS_PER_HZ
            )
            raise ValueError(
                f"IF frequency must be in [-{min_max_frequency_in_hz:e}, {min_max_frequency_in_hz:e}] Hz. "
                f"Got {freq_hz:e} Hz"
            )

        return steps

    def amp_as_steps(self, amp: float) -> int:
        """
        The instruction `set_awg_offs` expects DAC ratio as a (potentially signed) integer operand.
        """

        amp_steps = int(amp.real * Constants.MAX_OFFSET)
        if amp_steps < Constants.MIN_OFFSET or amp_steps > Constants.MAX_OFFSET:
            raise ValueError(
                f"""
                Expected offset to be in range [{Constants.MIN_OFFSET}, {Constants.MAX_OFFSET}].
                Got {amp_steps} instead
                """
            )

        return amp_steps

    def _legalise_bound(self, name: str, bound: IterBound, inst: Instruction):
        legal_bound = bound

        if isinstance(inst, Acquire):
            num_bins = bound.count
            if num_bins > Constants.MAX_012_BINNED_ACQUISITIONS:
                raise ValueError(
                    f"""
                        Loop nest size would require {num_bins} acquisition memory bins which exceeds the maximum {Constants.MAX_012_BINNED_ACQUISITIONS}.
                        Please reduce number of points
                        """
                )
        elif isinstance(inst, Delay):
            if bound.start < Constants.GRID_TIME:
                log.warning(
                    f"Undefined runtime behaviour. Variable {name} has illegal lower bound"
                )
            if bound.end > Constants.MAX_WAIT_TIME:
                log.warning(
                    f"Undefined runtime behaviour. Will be batching variable {name} at runtime"
                )
        elif isinstance(inst, DeviceUpdate):
            if inst.attribute not in ["frequency", "phase"]:
                raise NotImplementedError(
                    f"Unsupported processing of attribute {inst.attribute} for instruction {inst}"
                )

            if inst.attribute == "frequency":
                legal_bound = IterBound(
                    start=self.freq_as_steps(bound.start),
                    step=self.freq_as_steps(bound.step),
                    end=self.freq_as_steps(bound.end),
                    count=bound.count,
                )
            elif inst.attribute == "phase":
                legal_bound = IterBound(
                    start=self.phase_as_steps(bound.start),
                    step=self.phase_as_steps(bound.step),
                    end=self.phase_as_steps(bound.end),
                    count=bound.count,
                )
        else:
            attr2var = {
                attr: var
                for attr, var in vars(inst).items()
                if isinstance(var, Variable) and var.name == name
            }

            if attr2var and isinstance(inst, Pulse) and inst.shape != PulseShapeType.SQUARE:
                raise ValueError("Cannot process non-trivial pulses")
            if not attr2var:
                return legal_bound
            if len(attr2var) > 1:
                raise ValueError(
                    f"Unsafe analysis. Distinct attributes expecting the same variable bound {attr2var}"
                )
            attr, var = next(iter(attr2var.items()), (None, None))
            if attr not in ["width", "amp", "phase"]:
                raise NotImplementedError(
                    f"Unsupported processing of attribute {attr} for instruction {inst}"
                )

            if attr == "width":
                if bound.start < Constants.GRID_TIME:
                    log.warning(
                        f"Undefined runtime behaviour. Variable {name} has illegal lower bound"
                    )
                if bound.end > Constants.MAX_WAIT_TIME:
                    log.warning(
                        f"Undefined runtime behaviour. Will be batching variable {name} at runtime"
                    )
            elif attr == "amp" and isinstance(inst, Pulse):
                legal_bound = IterBound(
                    start=self.amp_as_steps(bound.start),
                    step=self.amp_as_steps(bound.step),
                    end=self.amp_as_steps(bound.end),
                    count=bound.count,
                )
            elif attr == "phase":
                legal_bound = IterBound(
                    start=self.phase_as_steps(bound.start),
                    step=self.phase_as_steps(bound.step),
                    end=self.phase_as_steps(bound.end),
                    count=bound.count,
                )

        for attr, val in vars(legal_bound).items():
            if not isinstance(val, int):
                raise ValueError(
                    f"Illegal value {val} for attribute {attr}. Expected value to be an integer"
                )

        # Qblox registers are unsigned 32bit integers.
        return legal_bound.astype(np.uint32)

    def run(self, ir: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs):
        """
        Performs target-dependent legalisation for QBlox.

            A) A repeat instruction with a very high repetition count is illegal because acquisition memory
        on a QBlox sequencer is limited. This requires optimal batching of the repeat instruction into maximally
        supported batches of smaller repeat counts.

        This pass does not do any batching. More features and adjustments will follow in future iterations.

            B) Previously processed variables such as frequencies, phases, and amplitudes still need digital conversion
        to a representation that's required by the QBlox ISA.

        + NCO's 1GHz frequency range by 4e9 steps:
            + [-500, 500] Mhz <=> [-2e9, 2e9] steps
            + 1 Hz            <=> 4 steps
        + NCO's 360° phase range by 1e9 steps:
            + 1e9    steps    <=> 2*pi rad
            + 125e6  steps    <=> pi/4 rad
        + Time and samples are quantified in nanoseconds
        + Amplitude depends on the type of the module:
            + [-1, 1]         <=> [-2.5, 2.5] V (for QCM)
            + [-1, 1]         <=> [-0.5, 0.5] V (for QRM)
        + AWG offset:
            + [-1, 1]         <=> [-32 768, 32 767]

        The last point is interesting as it requires knowledge of physical configuration of qubits and the modules
        they are wired to. This knowledge is typically found during execution and involving it early on would upset
        the rest of the compilation flow. In fact, it complicates this pass in particular, involves allocation
        concepts that should not be treated here, and promotes a monolithic compilation style. A temporary workaround
        is to simply assume the legality of amplitudes from the start whereby users are required to convert
        the desired voltages to the equivalent ratio AOT.

        This pass performs target-dependent conversion as described in part (B). More features and adjustments
        will follow in future iterations.
        """

        triage_result = res_mgr.lookup_by_type(TriageResult)
        binding_result = res_mgr.lookup_by_type(BindingResult)

        for target in triage_result.target_map:
            rw_result = binding_result.rw_results[target]
            bound_result = binding_result.iter_bound_results[target]
            legal_bound_result: Dict[str, IterBound] = deepcopy(bound_result)

            qblox_bounds: Dict[str, Set[IterBound]] = defaultdict(set)
            for name, instructions in rw_result.reads.items():
                for inst in instructions:
                    legal_bound = self._legalise_bound(name, bound_result[name], inst)
                    qblox_bounds[name].add(legal_bound)

            for name, bound in bound_result.items():
                if name in qblox_bounds:
                    bound_set = qblox_bounds[name]
                    if len(bound_set) > 1:
                        raise ValueError(
                            f"Ambiguous Qblox bounds for variable {name} in target {target}"
                        )
                    legal_bound_result[name] = next(iter(bound_set))

            # TODO: the proper way is to produce a new result and invalidate the old one
            binding_result.iter_bound_results[target] = legal_bound_result


@dataclass
class CFGResult(ResultInfoMixin):
    cfg: ControlFlowGraph = field(default_factory=lambda: ControlFlowGraph())


class CFGPass(AnalysisPass):
    def run(self, ir: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs):
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        :param res_mgr: The result manager to save the analysis results.
        """

        result = CFGResult()
        self._build_cfg(ir, result.cfg)
        res_mgr.add(result)

    def _build_cfg(self, builder: InstructionBuilder, cfg: ControlFlowGraph):
        """
        Recursively (re)discovers (new) header nodes and flow information.

        Supports Repeat, Sweep, EndRepeat, and EndSweep. More control flow and branching instructions
        will follow in the future once these foundations sink in and get stabilised in the codebase
        """

        flow = [(e.src.head(), e.dest.head()) for e in cfg.edges]
        headers = sorted([n.head() for n in cfg.nodes]) or [
            i
            for i, inst in enumerate(builder.instructions)
            if isinstance(inst, (Sweep, Repeat, EndSweep, EndRepeat))
        ]
        if headers[0] != 0:
            headers.insert(0, 0)

        next_flow = set(flow)
        next_headers = set(headers)
        for i, h in enumerate(headers):
            inst_at_h = builder.instructions[h]
            src = cfg.get_or_create_node(h)
            if isinstance(inst_at_h, Repeat):
                next_headers.add(h + 1)
                next_flow.add((h, h + 1))
                dest = cfg.get_or_create_node(h + 1)
                cfg.get_or_create_edge(src, dest)
            elif isinstance(inst_at_h, Sweep):
                next_headers.add(h + 1)
                next_flow.add((h, h + 1))
                dest = cfg.get_or_create_node(h + 1)
                cfg.get_or_create_edge(src, dest)
            elif isinstance(inst_at_h, (EndSweep, EndRepeat)):
                if h < len(builder.instructions) - 1:
                    next_headers.add(h + 1)
                    next_flow.add((h, h + 1))
                    dest = cfg.get_or_create_node(h + 1)
                    cfg.get_or_create_edge(src, dest)
                delimiter_type = Sweep if isinstance(inst_at_h, EndSweep) else Repeat
                p = next(
                    (
                        p
                        for p in headers[i::-1]
                        if isinstance(builder.instructions[p], delimiter_type)
                    )
                )
                next_headers.add(p)
                next_flow.add((h, p))
                dest = cfg.get_or_create_node(p)
                cfg.get_or_create_edge(src, dest)
            else:
                k = next(
                    (
                        s
                        for s, inst in enumerate(builder.instructions[h + 1 :])
                        if isinstance(inst, (Sweep, Repeat, EndSweep, EndRepeat))
                    ),
                    None,
                )
                if k:
                    next_headers.add(k + h + 1)
                    next_flow.add((h, k + h + 1))
                    src.elements.extend(range(src.tail() + 1, k + h + 1))
                    dest = cfg.get_or_create_node(k + h + 1)
                    cfg.get_or_create_edge(src, dest)

        if next_headers == set(headers) and next_flow == set(flow):
            # TODO - In-place use instructions instead of indices directly
            for n in cfg.nodes:
                n.elements = [builder.instructions[i] for i in n.elements]
            return

        self._build_cfg(builder, cfg)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd
import itertools
from collections import defaultdict
from copy import deepcopy
from dataclasses import dataclass, field
from functools import reduce
from operator import mul

import numpy as np
from compiler_config.config import InlineResultsProcessing

from qat.backend.graph import ControlFlowGraph
from qat.core.pass_base import AnalysisPass, ResultManager
from qat.core.result_base import ResultInfoMixin
from qat.executables import AcquireData
from qat.ir.lowered import PartitionedIR
from qat.ir.measure import PostProcessing as PydPostProcessing
from qat.purr.backends.utilities import UPCONVERT_SIGN
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.devices import PhysicalChannel, PulseChannel, PulseShapeType
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.compiler.instructions import (
    Acquire,
    Assign,
    Delay,
    DeviceUpdate,
    EndRepeat,
    EndSweep,
    Instruction,
    PostProcessing,
    PostProcessType,
    ProcessAxis,
    Pulse,
    QuantumInstruction,
    Repeat,
    ResultsProcessing,
    Return,
    Sweep,
    Variable,
    calculate_duration,
)
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


@dataclass
class TriageResult(ResultInfoMixin):
    sweeps: list[Sweep] = field(default_factory=list)
    repeats: list[Repeat] = field(default_factory=list)
    returns: list[Return] = field(default_factory=list)
    assigns: list[Assign] = field(default_factory=list)
    target_map: dict[PulseChannel, list[Instruction]] = field(
        default_factory=lambda: defaultdict(list)
    )
    device_updates: list[DeviceUpdate] = field(default_factory=list)
    quantum_instructions: list[QuantumInstruction] = field(default_factory=list)
    acquire_map: dict[PulseChannel, list[Acquire]] = field(
        default_factory=lambda: defaultdict(list)
    )
    pp_map: dict[str, list[PostProcessing]] = field(
        default_factory=lambda: defaultdict(list)
    )
    rp_map: dict[str, InlineResultsProcessing] = field(default_factory=dict)
    active_targets: set[PulseChannel] = field(default_factory=set)


class TriagePass(AnalysisPass):
    """
    Builds a view of instructions per quantum target AOT.

    Builds selections of instructions useful for subsequent analysis/transform passes, for
    code generation, and post-playback steps.

    This is equivalent to the :class:`QatFile` and simplifies the duration timeline creation
    in the legacy code.
    """

    def run(self, ir: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs):
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        :param res_mgr: The result manager to save the analysis results.
        """

        targets = set()
        for inst in ir.instructions:
            if isinstance(inst, QuantumInstruction):
                if isinstance(inst, PostProcessing):
                    for qt in inst.quantum_targets:
                        if isinstance(qt, Acquire):
                            targets.update(qt.quantum_targets)
                        else:
                            targets.add(qt)
                else:
                    targets.update(inst.quantum_targets)

        result = TriageResult()
        for inst in ir.instructions:
            ### Dissect by target ###
            if isinstance(inst, QuantumInstruction):
                result.quantum_instructions.append(inst)
                for qt in inst.quantum_targets:
                    if isinstance(qt, Acquire):
                        for aqt in qt.quantum_targets:
                            result.target_map[aqt].append(inst)
                    else:
                        result.target_map[qt].append(inst)
            elif isinstance(inst, Sweep | EndSweep | Repeat | EndRepeat):
                # TODO: Find a better way to handle scopes with the new instructions
                # Would a `ControlFlowInstruction` class be useful?
                # Are `Label` and `Jump` instructions needed here as well?
                for t in targets:
                    result.target_map[t].append(inst)

            ### Assortments ###
            if isinstance(inst, Sweep):
                result.sweeps.append(inst)
            elif isinstance(inst, Repeat):
                result.repeats.append(inst)
            elif isinstance(inst, DeviceUpdate):
                result.device_updates.append(inst)
            elif isinstance(inst, Return):
                result.returns.append(inst)
            elif isinstance(inst, Assign):
                result.assigns.append(inst)
            elif isinstance(inst, PostProcessing):
                result.pp_map[inst.output_variable].append(inst)
            elif isinstance(inst, ResultsProcessing):
                result.rp_map[inst.variable] = inst.results_processing

            ### Active targets ###
            if isinstance(inst, Acquire):
                for t in inst.quantum_targets:
                    result.active_targets.add(t)
                    result.acquire_map[t.full_id()].append(inst)
            elif isinstance(inst, Pulse):
                result.active_targets.add(inst.channel)

        # Assume that raw acquisitions are experiment results.
        acquires = list(itertools.chain(*result.acquire_map.values()))
        missing_results = {
            acq.output_variable
            for acq in acquires
            if acq.output_variable not in result.rp_map
        }
        for missing_var in missing_results:
            result.rp_map[missing_var] = InlineResultsProcessing.Experiment

        res_mgr.add(result)

        return ir


@dataclass
class ScopingResult:
    scope2symbols: dict[tuple[Instruction, Instruction | None], set[str]] = field(
        default_factory=lambda: defaultdict(set)
    )
    symbol2scopes: dict[str, list[tuple[Instruction, Instruction | None]]] = field(
        default_factory=lambda: defaultdict(list)
    )


@dataclass
class ReadWriteResult:
    inits: dict[str, list[Instruction]] = field(default_factory=lambda: defaultdict(list))
    reads: dict[str, list[Instruction]] = field(default_factory=lambda: defaultdict(list))
    writes: dict[str, list[Instruction]] = field(default_factory=lambda: defaultdict(list))


@dataclass(frozen=True)
class IterBound:
    start: int | float | complex = None
    step: int | float | complex = None
    end: int | float | complex = None
    count: int = None

    def astype(self, ty: type):
        return IterBound(
            start=np.array([self.start], dtype=type(self.start)).view(ty)[0],
            step=np.array([self.step], dtype=type(self.step)).view(ty)[0],
            end=np.array([self.end], dtype=type(self.end)).view(ty)[0],
            count=self.count,
        )


@dataclass
class BindingResult(ResultInfoMixin):
    scoping_results: dict[PulseChannel, ScopingResult] = field(
        default_factory=lambda: defaultdict(lambda: ScopingResult())
    )
    rw_results: dict[PulseChannel, ReadWriteResult] = field(
        default_factory=lambda: defaultdict(lambda: ReadWriteResult())
    )
    iter_bound_results: dict[PulseChannel, dict[str, IterBound]] = field(
        default_factory=lambda: defaultdict(dict)
    )

    # TODO - COMPILER-727 - Move acquire data analysis to its own pass
    # TODO - COMPILER-860 - Safer and more flexible acquisition addressing
    acquire_data_map: dict[str, AcquireData] = field(default_factory=dict)


class BindingPass(AnalysisPass):
    """
    Builds binding of variables, instructions, and a view of variables from/to scopes.

    Variables are implicitly declared in sweep instructions and are ultimately read from
    quantum instructions. Thus, every iteration variable is associated to all the scopes it
    is declared in.

    Values of the iteration variable are abstract and don't mean anything. In this pass, we
    only extract  the bound and associate it to the name variable. Further analysis is
    required on read sites to make sure their usage is consistent and meaningful.
    """

    @staticmethod
    def extract_iter_bound(value: list | np.ndarray):
        """
        Given a sequence of numbers (typically having been generated from
        :code:`np.linspace()`), figure out if the numbers are linearly/evenly spaced,
        in which case returns an IterBound instance holding the start, step, end, and count
        of the numbers in the array, or else fail.

        In the future, we might be interested in relaxing this condition and return
        "interpolated" evenly spaced approximation of the input sequence.
        """

        if value is None:
            raise ValueError(f"Cannot process value {value}")

        if isinstance(value, np.ndarray):
            value = value.tolist()

        if not value:
            raise ValueError(f"Cannot process value {value}")

        start = value[0]
        step = 0
        end = value[-1]
        count = len(value)

        if count < 2:
            return IterBound(start, step, end, count)

        step = value[1] - value[0]

        if not np.isclose(step, (end - start) / (count - 1)):
            raise ValueError(f"Not a regularly partitioned space {value}")

        return IterBound(start, step, end, count)

    def run(self, ir: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs):
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        :param res_mgr: The result manager to save the analysis results.
        """

        # TODO -COMPILER-851 - Specify ctrl HW features in TargetData
        enable_hw_averaging = kwargs.get("enable_hw_averaging", False)

        # TODO - COMPILER-855 - Split into smaller passes
        triage_result = res_mgr.lookup_by_type(TriageResult)
        result = BindingResult()
        for target, instructions in triage_result.target_map.items():
            scoping_result = result.scoping_results[target]
            rw_result = result.rw_results[target]
            iter_bound_result = result.iter_bound_results[target]

            stack = []
            for inst in instructions:
                parent_scopes = [
                    (s, e) for (s, e) in scoping_result.scope2symbols if s in stack
                ]
                if isinstance(inst, Sweep):
                    stack.append(inst)
                    scope = (inst, None)
                    if p := next(iter(parent_scopes[::-1]), None):
                        for name in scoping_result.scope2symbols[p]:
                            if name not in scoping_result.scope2symbols[scope]:
                                scoping_result.scope2symbols[scope].add(name)
                                scoping_result.symbol2scopes[name].append(scope)

                    for name, value in inst.variables.items():
                        scoping_result.scope2symbols[scope].add(name)
                        scoping_result.symbol2scopes[name].append(scope)
                        iter_bound_result[name] = self.extract_iter_bound(value)

                    iter_name = f"{hash(inst)}"
                    rw_result.inits[iter_name].append(inst)
                    rw_result.reads[iter_name].append(inst)
                    rw_result.writes[iter_name].append(inst)
                elif isinstance(inst, Repeat):
                    stack.append(inst)
                    scope = (inst, None)
                    if p := next(iter(parent_scopes[::-1]), None):
                        for name in scoping_result.scope2symbols[p]:
                            if name not in scoping_result.scope2symbols[scope]:
                                scoping_result.scope2symbols[scope].add(name)
                                scoping_result.symbol2scopes[name].append(scope)

                    iter_name = f"{hash(inst)}"
                    scoping_result.scope2symbols[scope].add(iter_name)
                    scoping_result.symbol2scopes[iter_name].append(scope)
                    iter_bound_result[iter_name] = IterBound(
                        start=1, step=1, end=inst.repeat_count, count=inst.repeat_count
                    )

                    iter_name = f"{hash(inst)}"
                    rw_result.inits[iter_name].append(inst)
                    rw_result.reads[iter_name].append(inst)
                    rw_result.writes[iter_name].append(inst)
                elif isinstance(inst, (EndSweep, EndRepeat)):
                    try:
                        delimiter = stack.pop()
                    except IndexError:
                        raise ValueError(f"Unbalanced scope. Found orphan {inst}")

                    delimiter_type = Sweep if isinstance(inst, EndSweep) else Repeat
                    if not isinstance(delimiter, delimiter_type):
                        raise ValueError(f"Unbalanced scope. Found orphan {inst}")

                    scope = next(
                        (
                            (s, e)
                            for (s, e) in scoping_result.scope2symbols.keys()
                            if s == delimiter
                        )
                    )
                    symbols = scoping_result.scope2symbols[scope]
                    del scoping_result.scope2symbols[scope]
                    scoping_result.scope2symbols[(delimiter, inst)] = symbols

                    for name, scopes in scoping_result.symbol2scopes.items():
                        scoping_result.symbol2scopes[name] = [
                            (delimiter, inst) if s == scope else s for s in scopes
                        ]
                elif isinstance(inst, Acquire):
                    if enable_hw_averaging and any(
                        pp.process == PostProcessType.MEAN
                        and ProcessAxis.SEQUENCE in pp.axes
                        for pp in triage_result.pp_map[inst.output_variable]
                    ):
                        # Discard scopes defined by Repeat instructions
                        parent_scopes = [
                            (s, e) for (s, e) in parent_scopes if not isinstance(s, Repeat)
                        ]

                    iter_name = f"{hash(inst)}"
                    shape = tuple(
                        iter_bound_result[f"{hash(s)}"].count for (s, e) in parent_scopes
                    )
                    loop_nest_size = reduce(mul, shape, 1)
                    iter_bound_result[iter_name] = IterBound(
                        start=0, step=1, end=loop_nest_size, count=loop_nest_size
                    )

                    # TODO - COMPILER-727 - Move acquire data analysis to its own pass
                    # TODO - COMPILER-860 - Safer and more flexible acquisition addressing
                    if inst.output_variable in result.acquire_data_map:
                        raise ValueError(f"Key {inst.output_variable} already exists")

                    # TODO - COMPILER-852 - Contractual data specification
                    if len(parent_scopes) == 0:
                        assert shape == ()
                        shape = (1, -1)
                    elif not any(
                        (s, e) for (s, e) in parent_scopes if isinstance(s, Sweep)
                    ):
                        shape = (1,) + shape

                    legacy_pps = triage_result.pp_map[inst.output_variable]
                    result.acquire_data_map[inst.output_variable] = AcquireData(
                        mode=inst.mode,
                        shape=shape,
                        physical_channel=target.physical_channel.id,
                        post_processing=[
                            PydPostProcessing._from_legacy(pp) for pp in legacy_pps
                        ],
                        results_processing=triage_result.rp_map.get(
                            inst.output_variable, None
                        ),
                    )

                    iter_name = f"{hash(inst)}"
                    rw_result.reads[iter_name].append(inst)
                    if (
                        innermost := next((s for (s, e) in parent_scopes[::-1]), None)
                    ) is not None:
                        rw_result.writes[iter_name].append(innermost)
                else:
                    attr2var = {
                        attr: var
                        for attr, var in vars(inst).items()
                        if isinstance(var, Variable)
                    }

                    for attr, var in attr2var.items():
                        defining_sweep = next(
                            (
                                s
                                for s in stack[::-1]
                                if isinstance(s, Sweep) and var.name in s.variables
                            ),
                            None,
                        )
                        if defining_sweep is None:
                            raise ValueError(
                                f"Variable {var} referenced but no prior declaration found in target {target}"
                            )

                        rw_result.inits[var.name].append(defining_sweep)
                        rw_result.reads[var.name].append(inst)
                        rw_result.writes[var.name].append(defining_sweep)

            if len(stack) > 0:
                raise ValueError(
                    f"Unbalanced scopes. Found orphans {stack} in target {target}"
                )

        res_mgr.add(result)
        return ir


class TILegalisationPass(AnalysisPass):
    """
    An instruction is legal if it has a direct equivalent in the programming model
    implemented by the control stack. The notion of "legal" is highly determined by the
    hardware features of the control stack as well as its programming model. Control stacks
    such as Qblox have a direct ISA-level representation for basic RF instructions such as
    frequency and phase manipulation, arithmetic instructions such as add,  and branching
    instructions such as jump.

    This pass performs target-independent legalisation. The goal here is to understand how
    variables are used and legalise their bounds. Furthermore, analysis in this pass is
    fundamentally based on QAT semantics and must be kept target-agnostic so that it can be
    reused among backends.

    Particularly in QAT:
    #. A sweep instruction is illegal because it specifies unclear iteration semantics.
    #. Device updates/assigns in general are illegal because they are bound to a sweep
       instruction via a variable. In fact, a variable (implicitly defined by a Sweep
       instruction) remains obscure until a "read" (usually on the instruction builder or on
       the hardware model) (typically from a DeviceUpdate instruction) is encountered where
       its intent becomes clear. We say that a DeviceUpdate carries meaning for the variable
       and materialises its intention.
    """

    @staticmethod
    def decompose_freq(frequency: float, target: PulseChannel):
        if target.fixed_if:  # NCO freq constant
            nco_freq = target.baseband_if_frequency
            lo_freq = frequency - nco_freq
        else:  # LO freq constant
            lo_freq = target.baseband_frequency
            nco_freq = frequency - lo_freq

        return lo_freq, nco_freq

    @staticmethod
    def transform_amp(amp: float, scale_factor: float, ignore_scale, target: PulseChannel):
        bias = target.bias
        scale = (1.0 if ignore_scale else target.scale) + 0.0j
        pulse_amp = scale * (scale_factor * amp) + bias

        if abs(pulse_amp.real) > 1 or abs(pulse_amp.imag) > 1:
            raise ValueError("Illegal DAC/ADC ratio. It must be within range [-1, 1]")

        if pulse_amp.imag != 0:
            raise NotImplementedError("Unsupported processing of complex amplitudes")

        return pulse_amp.real

    def _legalise_bound(self, name: str, bound: IterBound, inst: Instruction):
        legal_bound = bound

        if isinstance(inst, Delay):
            legal_bound = IterBound(
                start=int(calculate_duration(Delay(inst.quantum_targets, bound.start))),
                step=int(calculate_duration(Delay(inst.quantum_targets, bound.step))),
                end=int(calculate_duration(Delay(inst.quantum_targets, bound.end))),
                count=bound.count,
            )
        elif isinstance(inst, DeviceUpdate):
            if inst.attribute not in ["frequency", "scale", "amp"]:
                raise NotImplementedError(
                    f"Unsupported processing of attribute {inst.attribute} for instruction {inst}"
                )

            if inst.attribute == "frequency":
                if inst.target.fixed_if:
                    raise ValueError(
                        f"fixed_if must be False on target {inst.target} to sweep over frequencies"
                    )
                legal_bound = IterBound(
                    start=self.decompose_freq(bound.start, inst.target)[1],
                    step=bound.step,
                    end=self.decompose_freq(bound.end, inst.target)[1],
                    count=bound.count,
                )
        else:
            attr2var = {
                attr: var
                for attr, var in vars(inst).items()
                if isinstance(var, Variable) and var.name == name
            }

            if not attr2var:
                return legal_bound
            if len(attr2var) > 1:
                raise ValueError(
                    f"Unsafe analysis. Distinct attributes expecting the same variable bound {attr2var}"
                )
            if isinstance(inst, Pulse) and inst.shape != PulseShapeType.SQUARE:
                raise ValueError("Cannot process non-trivial pulses")
            attr, var = next(iter(attr2var.items()))
            if attr not in ["width", "amp", "phase"]:
                raise NotImplementedError(
                    f"Unsupported processing of attribute {attr} for instruction {inst}"
                )

            target: PulseChannel = next(iter(inst.quantum_targets))
            if attr == "width":
                legal_bound = IterBound(
                    start=int(calculate_duration(Delay(target, bound.start))),
                    step=int(calculate_duration(Delay(target, bound.step))),
                    end=int(calculate_duration(Delay(target, bound.end))),
                    count=bound.count,
                )
            elif attr == "amp" and isinstance(inst, Pulse):
                legal_bound = IterBound(
                    start=self.transform_amp(
                        bound.start,
                        inst.scale_factor,
                        inst.ignore_channel_scale,
                        target,
                    ),
                    step=self.transform_amp(
                        bound.step, inst.scale_factor, inst.ignore_channel_scale, target
                    ),
                    end=self.transform_amp(
                        bound.end, inst.scale_factor, inst.ignore_channel_scale, target
                    ),
                    count=bound.count,
                )

        return legal_bound

    def run(self, ir: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs):
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        :param res_mgr: The result manager to save the analysis results.
        """

        triage_result = res_mgr.lookup_by_type(TriageResult)
        binding_result = res_mgr.lookup_by_type(BindingResult)

        for target in triage_result.target_map:
            rw_result = binding_result.rw_results[target]
            bound_result = binding_result.iter_bound_results[target]
            legal_bound_result: dict[str, IterBound] = deepcopy(bound_result)

            read_bounds: dict[str, set[IterBound]] = defaultdict(set)
            for name, instructions in rw_result.reads.items():
                for inst in instructions:
                    legal_bound = self._legalise_bound(name, bound_result[name], inst)
                    read_bounds[name].add(legal_bound)

            for name, bound_set in read_bounds.items():
                if len(bound_set) > 1:
                    raise ValueError(
                        f"Ambiguous bounds for variable {name} in target {target}"
                    )
                legal_bound_result[name] = next(iter(bound_set))

            # TODO: the proper way is to produce a new result and invalidate the old one
            binding_result.iter_bound_results[target] = legal_bound_result
        return ir


@dataclass
class CFGResult(ResultInfoMixin):
    cfg: ControlFlowGraph = field(default_factory=lambda: ControlFlowGraph())


class CFGPass(AnalysisPass):
    def run(self, ir: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs):
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        :param res_mgr: The result manager to save the analysis results.
        """

        result = CFGResult()
        self._build_cfg(ir, result.cfg)
        res_mgr.add(result)
        return ir

    def _build_cfg(self, builder: InstructionBuilder, cfg: ControlFlowGraph):
        """
        Recursively (re)discovers (new) header nodes and flow information.

        Supports Repeat, Sweep, EndRepeat, and EndSweep. More control flow and branching instructions
        will follow in the future once these foundations sink in and get stabilised in the codebase
        """

        flow = [(e.src.head(), e.dest.head()) for e in cfg.edges]
        headers = sorted([n.head() for n in cfg.nodes]) or [
            i
            for i, inst in enumerate(builder.instructions)
            if isinstance(inst, (Sweep, Repeat, EndSweep, EndRepeat))
        ]
        if headers[0] != 0:
            headers.insert(0, 0)

        next_flow = set(flow)
        next_headers = set(headers)
        for i, h in enumerate(headers):
            inst_at_h = builder.instructions[h]
            src = cfg.get_or_create_node(h)
            if isinstance(inst_at_h, Repeat):
                next_headers.add(h + 1)
                next_flow.add((h, h + 1))
                dest = cfg.get_or_create_node(h + 1)
                cfg.get_or_create_edge(src, dest)
            elif isinstance(inst_at_h, Sweep):
                next_headers.add(h + 1)
                next_flow.add((h, h + 1))
                dest = cfg.get_or_create_node(h + 1)
                cfg.get_or_create_edge(src, dest)
            elif isinstance(inst_at_h, (EndSweep, EndRepeat)):
                if h < len(builder.instructions) - 1:
                    next_headers.add(h + 1)
                    next_flow.add((h, h + 1))
                    dest = cfg.get_or_create_node(h + 1)
                    cfg.get_or_create_edge(src, dest)
                delimiter_type = Sweep if isinstance(inst_at_h, EndSweep) else Repeat
                p = next(
                    (
                        p
                        for p in headers[i::-1]
                        if isinstance(builder.instructions[p], delimiter_type)
                    )
                )
                next_headers.add(p)
                next_flow.add((h, p))
                dest = cfg.get_or_create_node(p)
                cfg.get_or_create_edge(src, dest)
            else:
                k = next(
                    (
                        s
                        for s, inst in enumerate(builder.instructions[h + 1 :])
                        if isinstance(inst, (Sweep, Repeat, EndSweep, EndRepeat))
                    ),
                    None,
                )
                if k:
                    next_headers.add(k + h + 1)
                    next_flow.add((h, k + h + 1))
                    src.elements.extend(range(src.tail() + 1, k + h + 1))
                    dest = cfg.get_or_create_node(k + h + 1)
                    cfg.get_or_create_edge(src, dest)

        if next_headers == set(headers) and next_flow == set(flow):
            # TODO - In-place use instructions instead of indices directly
            for n in cfg.nodes:
                n.elements = [builder.instructions[i] for i in n.elements]
            return

        self._build_cfg(builder, cfg)


class LifetimePass(AnalysisPass):
    """
    The end goal of this pass is to facilitate sequencer allocation on the control hardware. Much like classical
    register allocation techniques, this pass is intended to perform "channel liveness" analysis.

    A logical channel is alive at some point P1 in the builder (analogically to a classical program) if it is
    targeted by some quantum operation at some point P2 > P1 in the future relative to P1.

    Example:

    P1:     |-- builder.pulse(pulse_channel1)
            |   ...
    P2:     |   builder.pulse(pulse_channel2) --|
            |   ...                             |
    P3:     |-- builder.pulse(pulse_channel1)   |
                ...                             |
    P4:         builder.pulse(pulse_channel2) --|

    pulse_channel1 is alive at P1 and P2, pulse_channel2 is alive at P2 and P3. Notice how the lifetimes of
    the channels overlap and "interfere".

    Knowledge of channel/target liveness (with full awareness of control flow) is invaluable for understanding
    physical allocation requirements on the control stack. This is achieved via an interference graph which
    allows allocation to be represented as a graph coloring.

    With this in mind, this pass spits out a colored interference graph that will be used by the code generator.
    """

    def run(self, ir: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs):
        pass


@dataclass
class PulseChannelTimeline:
    """Timeline analysis for instructions on a pulse channel.

    Imagine the timeline for a pulse channel, with an instruction that occurs over samples
    3-7, i.e.,

        samples: 0 1 2 [3 4 5 6 7] 8 9 10.

    The `start_position` would be 3, the `end_position` 7, and the number of `samples` 5.

    :param np.ndarray[int] samples: The number of samples each instruction takes.
    :param np.ndarray[int] start_positions: The sample when the instruction begins.
    :param np.ndarray[int] end_positions: The sample when the instruction ends.
    """

    samples: np.ndarray[int] = field(default_factory=lambda: np.ndarray([]))
    start_positions: np.ndarray[int] = field(default_factory=lambda: np.ndarray([]))
    end_positions: np.ndarray[int] = field(default_factory=lambda: np.ndarray([]))


@dataclass
class TimelineAnalysisResult(ResultInfoMixin):
    """Stores the timeline analysis for all pulse channels.

    :param target_map: The dictionary containing the timeline analysis for all pulse
        channels.
    """

    target_map: dict[PulseChannel, PulseChannelTimeline] = field(
        default_factory=lambda: defaultdict(PulseChannelTimeline)
    )
    total_duration: float = field(default=0.0)


class TimelineAnalysis(AnalysisPass):
    """Analyses the timeline of each pulse channel.

    Takes the instruction list for each pulse channel retrieved from the the partitioned
    results, and calculates the timeline in units of samples (each sample takes time
    `sample_time`). It calculates the duration of each instruction in units of samples,
    and the start and end times of each instruction in units of samples.

    .. warning::

        The pass will assume that the durations of instructions are sanitised to the
        granularity of the channels. If instructions that do not meet the criteria are
        provided, it might produce incorrect timelines. This can be enforced used the
        :class:`InstructionGranularitySanitisation <qat.middleend.passes.transform.InstructionGranularitySanitisation>`
        pass.
    """

    def run(
        self, ir: PartitionedIR, res_mgr: ResultManager, *args, **kwargs
    ) -> PartitionedIR:
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        :param res_mgr: The result manager to store the analysis results.
        """

        target_map = ir.target_map

        result = TimelineAnalysisResult()
        total_duration = 0
        for pulse_channel, instructions in target_map.items():
            pulse_channel_durations = np.array(
                [
                    (inst.duration if isinstance(inst, QuantumInstruction) else 0.0)
                    for inst in instructions
                ]
            )
            total_duration = max(total_duration, np.sum(pulse_channel_durations))

            durations = self.durations_as_samples(pulse_channel, pulse_channel_durations)
            cumulative_durations = np.cumsum(durations)
            result.target_map[pulse_channel] = PulseChannelTimeline(
                samples=durations,
                start_positions=cumulative_durations - durations,
                end_positions=cumulative_durations - 1,
            )

        result.total_duration = total_duration
        res_mgr.add(result)
        return ir

    @staticmethod
    def durations_as_samples(channel: PulseChannel, durations: list[float]):
        """Converts a list of durations into a number of samples."""

        block_numbers = np.ceil(
            np.round(durations / channel.block_time, decimals=4)
        ).astype(np.int64)
        return block_numbers * channel.block_size


@dataclass
class IntermediateFrequencyResult(ResultInfoMixin):
    frequencies: dict[PhysicalChannel, float]


class IntermediateFrequencyAnalysis(AnalysisPass):
    """
    Adapted from :meth:`qat.purr.backends.live.LiveDeviceEngine.build_baseband_frequencies`.

    Retrieves intermediate frequencies for all physical channels if they exist,
    and validates that pulse channels that share the same physical channel cannot
    have differing fixed frequencies. This pass should always follow a :class:`TriagePass`,
    as information of pulse channels are needed.
    """

    def __init__(self, model: QuantumHardwareModel):
        """
        Instantiate the pass with a hardware model.

        :param model: The hardware model.
        """

        self.model = model

    def run(
        self, ir: PartitionedIR, res_mgr: ResultManager, *args, **kwargs
    ) -> PartitionedIR:
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        :param res_mgr: The result manager to store the analysis results.
        """

        baseband_freqs = {}
        baseband_freqs_fixed_if = {}
        for pulse_channel in ir.target_map:
            if pulse_channel.fixed_if:
                baseband_freq = (
                    pulse_channel.frequency
                    - UPCONVERT_SIGN * pulse_channel.baseband_if_frequency
                )
                if (
                    baseband_freqs_fixed_if.get(pulse_channel.physical_channel, False)
                    and baseband_freqs[pulse_channel.physical_channel] != baseband_freq
                ):
                    raise ValueError(
                        "Cannot fix the frequency for two pulse channels of different "
                        "frequencies on the same physical channel!"
                    )
                baseband_freqs[pulse_channel.physical_channel] = baseband_freq
                baseband_freqs_fixed_if[pulse_channel.physical_channel] = (
                    pulse_channel.fixed_if
                )
            else:
                if (
                    pulse_channel.physical_channel not in baseband_freqs_fixed_if
                    or not baseband_freqs_fixed_if[pulse_channel.physical_channel]
                ):
                    baseband_freqs_fixed_if[pulse_channel.physical_channel] = (
                        pulse_channel.fixed_if
                    )

        res_mgr.add(IntermediateFrequencyResult(frequencies=baseband_freqs))
        return ir



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd

from abc import ABC
from collections import defaultdict
from contextlib import ExitStack, contextmanager
from numbers import Number

import numpy as np
from more_itertools import partition

from qat.backend.base import AllocatingBackend
from qat.backend.codegen_base import DfsTraversal
from qat.backend.passes.purr.analysis import (
    BindingPass,
    CFGResult,
    ReadWriteResult,
    ScopingResult,
    TILegalisationPass,
    TriagePass,
)
from qat.backend.passes.purr.transform import ScopePeeling
from qat.backend.qblox.config.constants import Constants
from qat.backend.qblox.config.specification import ModuleConfig, SequencerConfig
from qat.backend.qblox.execution import QbloxPackage, QbloxProgram
from qat.backend.qblox.ir import Opcode, SequenceBuilder
from qat.backend.qblox.passes.analysis import (
    AllocationManager,
    BindingResult,
    IterBound,
    PreCodegenResult,
    QbloxLegalisationPass,
    TriageResult,
)
from qat.core.metrics_base import MetricsManager
from qat.core.pass_base import PassManager
from qat.core.result_base import ResultManager
from qat.executables import Executable
from qat.purr.backends.utilities import evaluate_shape
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.devices import PulseChannel, PulseShapeType
from qat.purr.compiler.execution import DeviceInjectors, SweepIterator
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.compiler.instructions import (
    Acquire,
    CustomPulse,
    Delay,
    DeviceUpdate,
    FrequencyShift,
    Id,
    Instruction,
    MeasurePulse,
    PhaseReset,
    PhaseShift,
    PostProcessing,
    Pulse,
    QuantumInstruction,
    Repeat,
    Sweep,
    Synchronize,
    Variable,
    Waveform,
    calculate_duration,
)
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class QbloxContext(ABC):
    def __init__(
        self,
        rw_result: ReadWriteResult,
        iter_bounds: dict[str, IterBound],
        alloc_mgr: AllocationManager = None,
    ):
        self.inits: dict[str, list[Instruction]] = rw_result.inits
        self.reads: dict[str, list[Instruction]] = rw_result.reads
        self.writes: dict[str, list[Instruction]] = rw_result.writes
        self.iter_bounds = iter_bounds
        self.alloc_mgr = alloc_mgr or AllocationManager()

        self.sequence_builder = SequenceBuilder()
        self.sequencer_config = SequencerConfig()

        self._num_hw_avg = 1  # Technically disabled
        self._wf_memory: int = Constants.MAX_SAMPLE_SIZE_WAVEFORMS
        self._wf_index: int = 0
        self._acq_index: int = 0
        self._wgt_index: int = 0

        self._durations: list[float | str] = []
        self._frequency: float = 0.0  # Tracks frequency shifts on the target
        self._phases: list[float | str] = []
        self._timeline: np.ndarray = np.empty(0, dtype=complex)

    def __enter__(self):
        """
        Serves as the prologue
        """

        self.clear()

        self.sequence_builder.set_mrk(3)
        self.sequence_builder.set_latch_en(1, 4)
        self.sequence_builder.upd_param(Constants.GRID_TIME)

        for name, register in self.alloc_mgr.registers.items():
            self.sequence_builder.move(
                0,
                register,
                f"Precautionary initialisation for variable {name}",
            )

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Serves as the epilogue
        """

        self.sequence_builder.stop()

    @property
    def durations(self):
        return self._durations

    @property
    def duration(self):
        if all((isinstance(d, Number) for d in self._durations)):
            return sum(self._durations)

        ValueError("Cannot determine duration statically in dynamic settings")

    @property
    def phase(self):
        if all((isinstance(p, Number) for p in self._phases)):
            return sum(self._phases)

        ValueError("Cannot determine phase statically in dynamic settings")

    def create_package(
        self, target: PulseChannel, seq_idx: int, slot_idx: int
    ) -> QbloxPackage:
        lo_freq, nco_freq = TILegalisationPass.decompose_freq(target.frequency, target)

        qblox_config = target.physical_channel.config

        # Customise Module config
        mod_config: ModuleConfig = qblox_config.module
        if mod_config.lo.out0_en:
            mod_config.lo.out0_freq = lo_freq
        if mod_config.lo.out1_en:
            mod_config.lo.out1_freq = lo_freq
        if mod_config.lo.out0_in0_en:
            mod_config.lo.out0_in0_freq = lo_freq

        mod_config.scope_acq.sequencer_select = seq_idx
        mod_config.scope_acq.trigger_mode_path0 = "sequencer"
        mod_config.scope_acq.avg_mode_en_path0 = True
        mod_config.scope_acq.trigger_mode_path1 = "sequencer"
        mod_config.scope_acq.avg_mode_en_path1 = True

        # Customise Sequencer config
        seq_config: SequencerConfig = qblox_config.sequencers[seq_idx]
        seq_config.nco.freq = nco_freq
        seq_config.nco.prop_delay_comp_en = True
        seq_config.square_weight_acq.integration_length = (
            self.sequencer_config.square_weight_acq.integration_length
        )
        seq_config.thresholded_acq.rotation = self.sequencer_config.thresholded_acq.rotation
        seq_config.thresholded_acq.threshold = (
            self.sequencer_config.thresholded_acq.threshold
        )

        return QbloxPackage(
            pulse_channel_id=target.full_id(),
            physical_channel_id=target.physical_channel_id,
            seq_idx=seq_idx,
            seq_config=seq_config,
            slot_idx=slot_idx,
            mod_config=mod_config,
            sequence=self.sequence_builder.build(),
            timeline=self._timeline,
        )

    def is_empty(self):
        """
        Masks away yet-to-be-supported second-state, cancellation, and cross cancellation targets
        This is temporary and criteria will change with more features coming in
        """
        return not (
            self.sequence_builder.waveforms
            or self.sequence_builder.acquisitions
            or [
                inst
                for inst in self.sequence_builder.q1asm_instructions
                if inst.opcode
                in [
                    Opcode.PLAY,
                    Opcode.SET_AWG_OFFSET,
                    Opcode.SET_AWG_GAIN,
                    Opcode.ACQUIRE,
                    Opcode.ACQUIRE_TTL,
                    Opcode.ACQUIRE_WEIGHED,
                ]
            ]
        )

    def clear(self):
        self.sequence_builder.waveforms.clear()
        self.sequence_builder.acquisitions.clear()
        self.sequence_builder.weights.clear()
        self.sequence_builder.q1asm_instructions.clear()

    def wait_imm(self, duration: int):
        """
        Waits for `duration` nanoseconds expressed as an immediate.

        `duration` must be positive
        """

        if duration <= 0:
            log.debug(f"Expected positive duration, but got {duration}")
            return

        quotient = duration // Constants.MAX_WAIT_TIME
        remainder = duration % Constants.MAX_WAIT_TIME

        if remainder < Constants.GRID_TIME:
            log.debug(f"Rounding up {remainder} ns to {Constants.GRID_TIME} ns")
            remainder = Constants.GRID_TIME
            duration = quotient * Constants.MAX_WAIT_TIME + remainder

        if quotient >= Constants.LOOP_UNROLL_THRESHOLD:
            with self._loop("wait_label", quotient):
                self.sequence_builder.wait(Constants.MAX_WAIT_TIME)
        elif quotient >= 1:
            for _ in range(quotient):
                self.sequence_builder.wait(Constants.MAX_WAIT_TIME)

        self.sequence_builder.wait(remainder)

    def wait_reg(self, duration: str):
        """
        A mere `wait RX` in general has undefined runtime behaviour.

        This is a useful helper to dynamically wait for a `duration` nanoseconds expressed as a register.
        `iter_reg` is a short-lived register only used as an interator. Customer is responsible for (de)allocation.
        """

        batch_enter = self.alloc_mgr.label_gen("batch_enter")
        batch_exit = self.alloc_mgr.label_gen("batch_exit")
        remainder = self.alloc_mgr.label_gen("remainder")
        round_up = self.alloc_mgr.label_gen("round_up")

        with self.alloc_mgr.reg_borrow("tmp") as iter_reg:
            if duration != iter_reg:
                self.sequence_builder.move(
                    duration,
                    iter_reg,
                    f"Avoids cluttering {duration} as it's likely used as an accumulator",
                )
                self.sequence_builder.nop()

            self.sequence_builder.jlt(iter_reg, Constants.GRID_TIME, round_up)
            self.sequence_builder.jlt(iter_reg, Constants.MAX_WAIT_TIME, remainder)
            self.sequence_builder.label(batch_enter)
            self.sequence_builder.wait(Constants.MAX_WAIT_TIME)
            self.sequence_builder.sub(iter_reg, Constants.MAX_WAIT_TIME, iter_reg)
            self.sequence_builder.nop()
            self.sequence_builder.jge(iter_reg, Constants.MAX_WAIT_TIME, batch_enter)
            self.sequence_builder.jge(iter_reg, Constants.GRID_TIME, remainder)
            self.sequence_builder.label(round_up)
            self.sequence_builder.jlt(iter_reg, 1, batch_exit)
            self.sequence_builder.move(Constants.GRID_TIME, iter_reg, "Rounding up")
            self.sequence_builder.nop()
            self.sequence_builder.label(remainder)
            self.sequence_builder.wait(iter_reg)
            self.sequence_builder.label(batch_exit)

    def ledger(self, duration: int | str, pulse: np.ndarray = None):
        """
        A helper method to keep track of the durations and phase as the codegen runs.

        Every little bit of time is accounted for and a timeline object (in static cases)
        is also kept up-to-date for debugging and visualisation purposes.
        """

        if isinstance(duration, int):
            pulse = np.zeros(duration, dtype=complex) if pulse is None else pulse
            if all((isinstance(p, Number) for p in self._phases)):
                total_phase = sum(self._phases)
                pulse = pulse * np.exp(1.0j * total_phase)
            assert pulse.size == duration
            self._durations.append(duration)
            if self._timeline is not None:
                self._timeline = np.append(self._timeline, pulse)
        elif isinstance(duration, str):
            self._durations.append(duration)
            self._timeline = None  # Destroy the timeline as we are in dynamic setting
        else:
            raise ValueError(f"Expected legal immediate or register, but got {duration}")

    @contextmanager
    def _loop(self, label: str, count: int = 1):
        """
        This context can be used to avoid unrolling loops but needs manual addition of the 0 iteration
        either immediately after the yield or immediately after the context has exited.
        """

        label = self.alloc_mgr.label_gen(label)
        register = self.alloc_mgr.reg_alloc(label)
        self.sequence_builder.move(count, register)
        self.sequence_builder.nop()
        self.sequence_builder.label(label)
        yield register
        self.sequence_builder.loop(register, label)
        self.alloc_mgr.reg_free(register)

    @contextmanager
    def _modulo_reg(self, a: str, n: int):
        """
        Helpful context that prepares a temporary register b such that a ≅ b [n] where
        n is some immediate constant batch size.
        """

        if not isinstance(n, int) or n < 2:
            raise ValueError(
                f"Invalid batch size. Expected an integer >= 2, but got {n} instead"
            )

        batch_enter = self.alloc_mgr.label_gen("batch_enter")
        batch_exit = self.alloc_mgr.label_gen("batch_exit")

        iter_reg = self.alloc_mgr.reg_alloc("tmp")
        if a != iter_reg:
            self.sequence_builder.move(
                a,
                iter_reg,
                f"Avoids cluttering {a} as it's likely used as an accumulator",
            )
            self.sequence_builder.nop()

        self.sequence_builder.jlt(iter_reg, n, batch_exit)
        self.sequence_builder.label(batch_enter)
        self.sequence_builder.sub(iter_reg, n, iter_reg)
        self.sequence_builder.nop()
        self.sequence_builder.jge(iter_reg, n, batch_enter)
        self.sequence_builder.label(batch_exit)
        yield iter_reg
        self.alloc_mgr.reg_free(iter_reg)

    def _upd_param_imm(self, duration: int):
        self.sequence_builder.upd_param(Constants.GRID_TIME)
        self.wait_imm(duration - Constants.GRID_TIME)

    def _upd_param_reg(self, duration: str):
        """
        Update latched parameters and wait for `duration` nanoseconds expressed as a register

        If `duration` is less than the threshold Constants.GRID_TIME, the generates assembly rounds up
        and waits for Constants.GRID_TIME.

        Note that the `up_param` instruction is defined only for an immediate operand which must be
        at least Constants.GRID_TIME. Therefore, technically we're already forced to round up.
        """

        batch_enter = self.alloc_mgr.label_gen("batch_enter")
        batch_exit = self.alloc_mgr.label_gen("batch_exit")
        remainder = self.alloc_mgr.label_gen("remainder")

        self.sequence_builder.upd_param(Constants.GRID_TIME)
        with self.alloc_mgr.reg_borrow("tmp") as iter_reg:
            self.sequence_builder.jlt(
                duration,
                Constants.GRID_TIME,
                batch_exit,
                "Guards against risky underflow",
            )
            self.sequence_builder.sub(duration, Constants.GRID_TIME, iter_reg)
            self.sequence_builder.nop()

            self.sequence_builder.jlt(iter_reg, Constants.GRID_TIME, batch_exit)
            self.sequence_builder.jlt(iter_reg, Constants.MAX_WAIT_TIME, remainder)
            self.sequence_builder.label(batch_enter)
            self.sequence_builder.wait(Constants.MAX_WAIT_TIME)
            self.sequence_builder.sub(iter_reg, Constants.MAX_WAIT_TIME, iter_reg)
            self.sequence_builder.nop()
            self.sequence_builder.jge(iter_reg, Constants.MAX_WAIT_TIME, batch_enter)
            self.sequence_builder.jlt(iter_reg, Constants.GRID_TIME, batch_exit)
            self.sequence_builder.label(remainder)
            self.sequence_builder.wait(iter_reg)
            self.sequence_builder.label(batch_exit)

    def _evaluate_waveform(self, waveform: Waveform, target: PulseChannel):
        """
        The waveform is evaluated as a 1d complex array. In QBlox, the Real and Imag parts of the pulse
        represent the digital offset on the AWG. They both must be within range [-1, 1].
        """

        num_samples = int(calculate_duration(waveform))
        dt = target.sample_time
        length = num_samples * dt
        centre = length / 2.0
        t = np.linspace(
            start=-centre + 0.5 * dt, stop=length - centre - 0.5 * dt, num=num_samples
        )
        pulse = evaluate_shape(waveform, t)
        scale = target.scale
        if isinstance(waveform, (Pulse, CustomPulse)) and waveform.ignore_channel_scale:
            scale = 1

        pulse *= scale
        pulse += target.bias

        if ((pulse.real < -1) | (pulse.real > 1)).any():
            raise ValueError(
                "Voltage range for I exceeded. Make sure all I values are within the range [-1, 1]"
            )
        if ((pulse.imag < -1) | (pulse.imag > 1)).any():
            raise ValueError(
                "Voltage range for Q exceeded. Make sure all I values are within the range [-1, 1]"
            )

        return pulse

    def _evaluate_square_pulse(self, pulse: Pulse, target: PulseChannel):
        bias = target.bias
        scale = 1.0 + 0.0j if pulse.ignore_channel_scale else target.scale

        pulse_amp = pulse.scale_factor * pulse.amp
        pulse_amp = scale * pulse_amp + bias

        if np.abs(pulse_amp.real) > 1 or np.abs(pulse_amp.imag) > 1:
            raise ValueError(
                f"""
                Voltage range for I or Q exceeded. Current values:
                waveform.scale_factor: {pulse.scale_factor}
                waveform.amp: {pulse.amp}
                (Channel) scale: {scale}
                (Channel) bias: {bias}
                """
            )

        return pulse_amp

    def _register_signal(self, waveform, target, data, name):
        index = self.sequence_builder.lookup_waveform_by_data(data)
        if index is not None:
            log.debug(
                f"Reusing signal {name} at index {index} for pulse {waveform} on channel {target}"
            )
        elif data.size > self._wf_memory:
            raise ValueError(
                f"No more waveform memory left for signal {name} of pulse {waveform} on channel {target}"
            )
        else:
            wf_hash = hash(waveform)
            wf_name = type(waveform).__name__
            if isinstance(waveform, Pulse):
                wf_name = waveform.shape.name

            index = self._wf_index
            self.sequence_builder.add_waveform(
                f"{wf_name}_{wf_hash}_{name}", index, data.tolist()
            )
            self._wf_memory = self._wf_memory - data.size
            self._wf_index = self._wf_index + 1

        return index

    def _register_waveform(self, waveform, target, data):
        i_index = self._register_signal(waveform, target, data.real, "I")
        q_index = self._register_signal(waveform, target, data.imag, "Q")
        return i_index, q_index

    def _register_acquisition(self, acq_name: str, num_bins: int | str):
        acq_index = self._acq_index
        self.sequence_builder.add_acquisition(acq_name, acq_index, num_bins)
        self._acq_index = acq_index + 1
        return acq_index

    def _register_weight(self, name: str, weight: np.ndarray):
        wgt_index = self._wgt_index
        self.sequence_builder.add_weight(name, wgt_index, weight.tolist())
        self._wgt_index = wgt_index + 1
        return wgt_index

    def _do_acquire(self, wgt_name, acq_filter, acq_index, acq_bin, acq_width):
        legal_acq_width = min(acq_width, Constants.MAX_WAIT_TIME)
        if acq_filter:
            weight = acq_filter.samples
            wgt_i_index = self._register_weight(f"{wgt_name}_weight_I", weight.real)
            wgt_q_index = self._register_weight(f"{wgt_name}_weight_Q", weight.imag)

            # TODO - silly convoluted way of using the `acquire_weighed` instruction
            # TODO - simplify when Qblox supports more freedom of memory addressing
            with (
                self.alloc_mgr.reg_borrow("wgt_i") as wgt_i_reg,
                self.alloc_mgr.reg_borrow("wgt_q") as wgt_q_reg,
            ):
                self.sequence_builder.move(wgt_i_index, wgt_i_reg)
                self.sequence_builder.move(wgt_q_index, wgt_q_reg)

                self.sequence_builder.acquire_weighed(
                    acq_index, acq_bin, wgt_i_reg, wgt_q_reg, legal_acq_width
                )
        else:
            self.sequence_builder.acquire(acq_index, acq_bin, legal_acq_width)

        self.wait_imm(acq_width - legal_acq_width)

    @contextmanager
    def _wrapper_pulse(
        self,
        delay_width,
        pulse_width,
        pulse_shape,
        i_steps,
        q_steps,
        i_index,
        q_index,
    ):
        effective_width = max(min(pulse_width, delay_width), Constants.GRID_TIME)
        if pulse_shape == PulseShapeType.SQUARE:
            self.sequence_builder.set_awg_offs(i_steps, q_steps)
            self.sequence_builder.upd_param(effective_width)
        else:
            self.sequence_builder.play(i_index, q_index, effective_width)
        remaining_width = pulse_width - effective_width
        if remaining_width < Constants.GRID_TIME:
            raise ValueError(
                f"""
                Remaining width after delay must be at least {Constants.GRID_TIME} ns.
                Got pulse width = {pulse_width} ns, delay width = {delay_width} ns. This leaves {remaining_width} ns,
                which isn't enough to trigger acquisition afterwards. Adjust the pulse width or delay width values so
                that the following holds:
                pulse_width - max(min(pulse_width, delay_width), Constants.GRID_TIME) >= Constants.GRID_TIME
                """
            )
        yield remaining_width
        if pulse_shape == PulseShapeType.SQUARE:
            self.sequence_builder.set_awg_offs(0, 0)
            self.sequence_builder.upd_param(Constants.GRID_TIME)

            self.ledger(Constants.GRID_TIME)

    @contextmanager
    def _wrapper_cond(self, mask, operator, duration):
        """
        A wrapper for conditional regions
        """

        self.sequence_builder.set_cond(
            1, mask, operator, duration, "Start of conditional region"
        )
        yield
        self.sequence_builder.set_cond(
            0, mask, operator, duration, "End of conditional region"
        )

    def id(self):
        self.sequence_builder.nop()

    def delay(self, inst: Delay):
        if isinstance(inst.duration, Variable):
            duration = self.alloc_mgr.registers[inst.duration.name]
            self.wait_reg(duration)
        elif isinstance(inst.duration, Number):
            duration = int(calculate_duration(inst))
            self.wait_imm(duration)
        else:
            raise ValueError(f"Expected a Variable or a Number but got {inst.duration}")

        self.ledger(duration)

    def shift_frequency(self, inst, target):
        # TODO - discuss w/t quantum_target.fixed_if is True or False
        old_frequency = target.frequency + self._frequency
        new_frequency = old_frequency + inst.frequency
        if new_frequency < target.min_frequency or new_frequency > target.max_frequency:
            raise ValueError(
                f"Cannot shift pulse channel frequency from '{old_frequency}' to '{new_frequency}'"
            )

        self._frequency = self._frequency + inst.frequency
        shifted_nco_freq = target.baseband_if_frequency + self._frequency
        value = QbloxLegalisationPass.freq_as_steps(shifted_nco_freq)  # 1 MHZ <-> 4e6
        self.sequence_builder.set_freq(value)
        self.sequence_builder.upd_param(Constants.GRID_TIME)

        self.ledger(Constants.GRID_TIME)

    def shift_phase(self, inst: PhaseShift):
        if isinstance(inst.phase, Variable):
            ph_reg = self.alloc_mgr.registers[inst.phase.name]
            with self._modulo_reg(ph_reg, Constants.NCO_MAX_PHASE_STEPS) as iter_reg:
                self.sequence_builder.set_ph_delta(iter_reg)
                self.sequence_builder.upd_param(Constants.GRID_TIME)
        elif isinstance(inst.phase, Number):
            ph_imm = QbloxLegalisationPass.phase_as_steps(inst.phase)
            self.sequence_builder.set_ph_delta(ph_imm)
            self.sequence_builder.upd_param(Constants.GRID_TIME)
        else:
            raise ValueError(f"Expected a Variable or a Number but got {inst.phase}")

        self._phases.append(inst.phase)
        self.ledger(Constants.GRID_TIME)

    def reset_phase(self):
        self.sequence_builder.reset_ph()
        self.sequence_builder.upd_param(Constants.GRID_TIME)

        self._phases.clear()
        self.ledger(Constants.GRID_TIME)

    def waveform(self, waveform: Waveform, target: PulseChannel):
        attr2var = {
            attr: var for attr, var in vars(waveform).items() if isinstance(var, Variable)
        }

        if isinstance(waveform, Pulse) and waveform.shape == PulseShapeType.SQUARE:
            if "width" in attr2var:
                pulse_width = attr2var["width"]
                pulse_width = self.alloc_mgr.registers[pulse_width.name]
                # TODO - Continue semantics: jump back to beginning of loop
            else:
                pulse_width = int(calculate_duration(waveform))
                if pulse_width < Constants.GRID_TIME:
                    log.debug(
                        f"""
                        Minimum pulse width is {Constants.GRID_TIME} ns with a resolution of {1} ns.
                        Please round up the width to at least {Constants.GRID_TIME} nanoseconds.
                        This pulse will be ignored.
                        """
                    )
                    return

            if "amp" in attr2var:
                pulse_amp = attr2var["amp"]
                i_steps = self.alloc_mgr.registers[pulse_amp.name]
                q_steps = self.alloc_mgr.registers["zero"]
            else:
                pulse_amp = self._evaluate_square_pulse(waveform, target)
                i_steps = int(pulse_amp.real * Constants.MAX_OFFSET)
                q_steps = int(pulse_amp.imag * Constants.MAX_OFFSET)

            self.sequence_builder.set_awg_offs(i_steps, q_steps)
            if "width" in attr2var:
                self._upd_param_reg(pulse_width)
            else:
                self._upd_param_imm(pulse_width)
            self.sequence_builder.set_awg_offs(0, 0)
            self.sequence_builder.upd_param(Constants.GRID_TIME)

            self.ledger(Constants.GRID_TIME)

            pulse = (
                np.ones(pulse_width, dtype=complex) * pulse_amp
                if isinstance(pulse_width, Number) and isinstance(pulse_amp, Number)
                else None
            )
        else:
            pulse = self._evaluate_waveform(waveform, target)
            pulse_width = pulse.size
            if pulse_width < Constants.GRID_TIME:
                log.debug(
                    f"""
                    Minimum pulse width is {Constants.GRID_TIME} ns with a resolution of {1} ns.
                    Please round up the width to at least {Constants.GRID_TIME} nanoseconds.
                    This pulse will be ignored.
                    """
                )
                return

            max_width = min(pulse_width, Constants.MAX_WAIT_TIME)
            i_index, q_index = self._register_waveform(waveform, target, pulse)
            self.sequence_builder.play(i_index, q_index, max_width)
            self.wait_imm(pulse_width - max_width)

        self.ledger(pulse_width, pulse)

    def measure_acquire(
        self, measure: MeasurePulse, acquire: Acquire, target: PulseChannel
    ):
        """
        Regarding HW thresholding, `acquire.rotation` is in radians and `acquire.threshold` is uncorrected.
        (Exactly as they've been computed during measure block construction)

        Here at configuration time, we "legalise" these parameters to what Qblox requires:
            + `rotation` as degrees
            + `threshold` to be corrected by the integration length
        """

        pulse = self._evaluate_waveform(measure, target)
        pulse_width = pulse.size
        delay_width = int(calculate_duration(Delay(acquire.channel, acquire.delay)))

        if pulse_width < Constants.GRID_TIME:
            log.debug(
                f"""
                Minimum pulse width is {Constants.GRID_TIME} ns with a resolution of {1} ns.
                Please round up the width to at least {Constants.GRID_TIME} nanoseconds.
                This pulse will be ignored. 
                """
            )
            return

        if pulse_width < delay_width + Constants.GRID_TIME:
            raise ValueError(
                f"""
                Expected pulse width >= delay width + {Constants.GRID_TIME} ns.
                Got pulse width = {pulse_width} ns and delay width = {delay_width} ns.
                """
            )

        name = f"{hash(acquire)}"
        num_bins = self.iter_bounds[name].count
        acq_bin = self.alloc_mgr.registers[name]
        acq_width = int(calculate_duration(acquire))
        acq_index = self._register_acquisition(acquire.output_variable, num_bins)
        self.sequencer_config.square_weight_acq.integration_length = acq_width
        self.sequencer_config.thresholded_acq.rotation = np.rad2deg(acquire.rotation)
        self.sequencer_config.thresholded_acq.threshold = acq_width * acquire.threshold

        i_steps, q_steps = None, None
        i_index, q_index = None, None
        if measure.shape == PulseShapeType.SQUARE:
            i_steps = int(pulse[0].real * Constants.MAX_OFFSET)
            q_steps = int(pulse[0].imag * Constants.MAX_OFFSET)
        else:
            i_index, q_index = self._register_waveform(measure, target, pulse)

        with self._wrapper_pulse(
            delay_width, pulse_width, measure.shape, i_steps, q_steps, i_index, q_index
        ) as remaining_width:
            self._do_acquire(
                acquire.output_variable,
                acquire.filter,
                acq_index,
                acq_bin,
                remaining_width,
            )

        self.ledger(pulse_width, pulse)

    @staticmethod
    def synchronize(inst: Synchronize, contexts: dict):
        """
        Favours static time padding whenever possible, or else uses SYNC.
        TODO - Default to SYNC when Qblox supports finer grained SYNC groups
        """

        is_static = all(
            all((isinstance(d, Number) for d in contexts[target].durations))
            for target in inst.quantum_targets
        )

        for target in inst.quantum_targets:
            cxt = contexts[target]
            if is_static:
                duration_offset = (
                    max((contexts[target].duration for target in inst.quantum_targets))
                    - cxt.duration
                )
                cxt.wait_imm(duration_offset)
                cxt.ledger(duration_offset)
            else:
                cxt.sequence_builder.wait_sync(Constants.GRID_TIME)
                cxt.ledger(Constants.GRID_TIME)

    def device_update(self, du_inst: DeviceUpdate):
        if isinstance(du_inst.value, Variable):
            val_reg = self.alloc_mgr.registers[du_inst.value.name]
            if du_inst.attribute == "frequency":
                self.sequence_builder.set_freq(val_reg)
                self.sequence_builder.upd_param(Constants.GRID_TIME)

                self.ledger(Constants.GRID_TIME)
            else:
                raise NotImplementedError(
                    f"Unsupported processing of attribute {du_inst.attribute} for instruction {du_inst}"
                )
        elif isinstance(du_inst.value, Number):
            if du_inst.attribute == "frequency":
                lo_freq, nco_freq = TILegalisationPass.decompose_freq(
                    du_inst.value, du_inst.target
                )
                freq_imm = QbloxLegalisationPass.freq_as_steps(nco_freq)
                self.sequence_builder.set_freq(freq_imm)
                self.sequence_builder.upd_param(Constants.GRID_TIME)

                self.ledger(Constants.GRID_TIME)
            elif du_inst.attribute == "scale":
                pass
            else:
                raise NotImplementedError(
                    f"Unsupported processing of attribute {du_inst.attribute} for instruction {du_inst}"
                )
        else:
            raise ValueError(f"Expected a Variable or a Number but got {du_inst.value}")

    @staticmethod
    def enter_repeat(inst: Repeat, contexts: dict):
        iter_name = f"{hash(inst)}"
        for context in contexts.values():
            var_names = set([n for (n, insts) in context.inits.items() if inst in insts])
            for name in var_names:
                register = context.alloc_mgr.registers[name]
                bound = context.iter_bounds[name]
                context.sequence_builder.move(
                    bound.start,
                    register,
                    f"Initialise register allocated for variable {name}",
                )

            label = context.alloc_mgr.labels[iter_name]
            context.sequence_builder.label(label)

            context.sequence_builder.wait_sync(
                Constants.GRID_TIME, "Sync at the beginning of shot"
            )
            context.reset_phase()

    @staticmethod
    def exit_repeat(inst: Repeat, contexts: dict):
        iter_name = f"{hash(inst)}"
        for context in contexts.values():
            context.wait_imm(int(inst.passive_reset_time * 1e9))

            var_names = set(
                [n for (n, insts) in context.writes.items() if inst in insts]
                + [n for (n, insts) in context.reads.items() if inst in insts]
            )
            for name in var_names:
                register = context.alloc_mgr.registers[name]
                bound = context.iter_bounds[name]
                context.sequence_builder.add(
                    register,
                    bound.step,
                    register,
                    f"Increment register for variable {name}",
                )
                context.sequence_builder.nop()

            register = context.alloc_mgr.registers[iter_name]
            bound = context.iter_bounds[iter_name]
            label = context.alloc_mgr.labels[iter_name]
            context.sequence_builder.jlt(register, bound.end + bound.step, label)

    @staticmethod
    def enter_sweep(inst: Sweep, contexts: dict):
        iter_name = f"{hash(inst)}"
        for context in contexts.values():
            var_names = set([n for (n, insts) in context.inits.items() if inst in insts])
            for name in var_names:
                register = context.alloc_mgr.registers[name]
                bound = context.iter_bounds[name]
                context.sequence_builder.move(
                    bound.start,
                    register,
                    f"Initialise register allocated for variable {name}",
                )

            label = context.alloc_mgr.labels[iter_name]
            context.sequence_builder.label(label)

    @staticmethod
    def exit_sweep(inst: Sweep, contexts: dict):
        iter_name = f"{hash(inst)}"
        for context in contexts.values():
            var_names = set(
                [n for (n, insts) in context.writes.items() if inst in insts]
                + [n for (n, insts) in context.reads.items() if inst in insts]
            )
            for name in var_names:
                register = context.alloc_mgr.registers[name]
                bound = context.iter_bounds[name]
                context.sequence_builder.add(
                    register,
                    bound.step,
                    register,
                    f"Increment register for variable {name}",
                )
                context.sequence_builder.nop()

            register = context.alloc_mgr.registers[iter_name]
            bound = context.iter_bounds[iter_name]
            label = context.alloc_mgr.labels[iter_name]
            context.sequence_builder.jlt(register, bound.end + bound.step, label)


class AbstractQbloxBackend(AllocatingBackend[QbloxProgram]):
    def __init__(self, model: QuantumHardwareModel, pipeline: PassManager = None):
        super().__init__(model=model, pipeline=pipeline)
        self.allocations: dict[int, dict[str, int]] = defaultdict(dict)

    def _do_emit(
        self,
        triage_result: TriageResult,
        binding_result: BindingResult,
        precodegen_result: PreCodegenResult,
        cfg_result: CFGResult,
        ignore_empty=True,
    ) -> QbloxProgram:
        rw_results: dict[PulseChannel, ReadWriteResult] = binding_result.rw_results
        iter_bound_results: dict[PulseChannel, dict[str, IterBound]] = (
            binding_result.iter_bound_results
        )
        alloc_mgrs: dict[PulseChannel, AllocationManager] = precodegen_result.alloc_mgrs

        with ExitStack() as stack:
            contexts = {
                t: stack.enter_context(
                    QbloxContext(
                        rw_result=rw_results[t],
                        iter_bounds=iter_bound_results[t],
                        alloc_mgr=alloc_mgrs[t],
                    )
                )
                for t in triage_result.target_map
            }
            QbloxCFGWalker(contexts).run(cfg_result.cfg)

        packages = {}
        for target, context in contexts.items():
            if not context.is_empty() or not ignore_empty:
                seq_idx, slot_idx = self.allocate(target)
                package = context.create_package(target, seq_idx, slot_idx)
                packages[target.full_id()] = package
        return QbloxProgram(packages=packages)


class QbloxBackend1(AbstractQbloxBackend):
    def emit(
        self,
        ir: InstructionBuilder,
        res_mgr: ResultManager,
        met_mgr: MetricsManager,
        ignore_empty: bool = True,
        **kwargs,
    ) -> Executable[QbloxProgram]:
        triage_result = res_mgr.lookup_by_type(TriageResult)
        device_updates = triage_result.device_updates
        quantum_instructions = triage_result.quantum_instructions
        sweeps = triage_result.sweeps
        switerator = SweepIterator.from_sweeps(sweeps)
        dinjectors = DeviceInjectors(device_updates)

        binding_result = res_mgr.lookup_by_type(BindingResult)
        scoping_result = next(
            iter(binding_result.scoping_results.values()), ScopingResult()
        )
        scopes = [(s, e) for (s, e) in scoping_result.scope2symbols if s in sweeps]

        ScopePeeling().run(ir, res_mgr, met_mgr, scopes=scopes)
        # TODO - Pass infrastructural improvements [COMPILER-849]
        # res_mgr.mark_as_dirty(triage_result, binding_result)

        programs: list[QbloxProgram] = []
        try:
            dinjectors.inject()
            while not switerator.is_finished():
                switerator.do_sweep(quantum_instructions)
                inner_res_mgr = ResultManager()
                inner_met_mgr = MetricsManager()
                (PassManager() | TriagePass() | BindingPass()).run(
                    ir, inner_res_mgr, inner_met_mgr, enable_hw_averaging=True
                )
                self.pipeline.run(ir, inner_res_mgr, inner_met_mgr, **kwargs)

                inner_triage_result = inner_res_mgr.lookup_by_type(TriageResult)
                inner_binding_result = inner_res_mgr.lookup_by_type(BindingResult)
                inner_precodegen_result = inner_res_mgr.lookup_by_type(PreCodegenResult)
                inner_cfg_result = inner_res_mgr.lookup_by_type(CFGResult)
                program = self._do_emit(
                    inner_triage_result,
                    inner_binding_result,
                    inner_precodegen_result,
                    inner_cfg_result,
                    ignore_empty,
                )
                programs.append(program)
        finally:
            switerator.revert(triage_result.quantum_instructions)
            dinjectors.revert()

        acquires = binding_result.acquire_data_map
        assigns = triage_result.assigns
        return_vars = set(next(iter(triage_result.returns)).variables)
        shots = (
            sum((repeat.repeat_count for repeat in triage_result.repeats), 0)
            if triage_result.repeats
            else None
        )

        return Executable[QbloxProgram](
            programs=programs,
            acquires=acquires,
            assigns=assigns,
            returns=return_vars,
            shots=shots,
            calibration_id=self.model.calibration_id,
        )


class QbloxBackend2(AbstractQbloxBackend):
    def emit(
        self,
        ir: InstructionBuilder,
        res_mgr: ResultManager,
        met_mgr: MetricsManager,
        ignore_empty: bool = True,
        **kwargs,
    ) -> Executable[QbloxProgram]:
        triage_result = res_mgr.lookup_by_type(TriageResult)
        device_updates = triage_result.device_updates

        remaining, static_dus = partition(
            lambda inst: inst in device_updates and not isinstance(inst.value, Variable),
            ir.instructions,
        )
        remaining, static_dus = list(remaining), list(static_dus)
        dinjectors = DeviceInjectors(static_dus)

        ir.instructions = remaining
        triage_result.device_updates = [
            inst for inst in device_updates if inst not in static_dus
        ]
        # TODO - Pass infrastructural improvements [COMPILER-849]
        # res_mgr.mark_as_dirty(triage_result)

        programs: list[QbloxProgram] = []
        try:
            dinjectors.inject()
            self.pipeline.run(ir, res_mgr, met_mgr, **kwargs)

            triage_result = res_mgr.lookup_by_type(TriageResult)
            binding_result = res_mgr.lookup_by_type(BindingResult)
            precodegen_result = res_mgr.lookup_by_type(PreCodegenResult)
            cfg_result = res_mgr.lookup_by_type(CFGResult)
            program = self._do_emit(
                triage_result,
                binding_result,
                precodegen_result,
                cfg_result,
                ignore_empty,
            )
            programs.append(program)
        finally:
            dinjectors.revert()

        acquires = binding_result.acquire_data_map
        assigns = triage_result.assigns
        return_vars = set(next(iter(triage_result.returns)).variables)
        shots = (
            sum((repeat.repeat_count for repeat in triage_result.repeats), 0)
            if triage_result.repeats
            else None
        )

        return Executable[QbloxProgram](
            programs=programs,
            acquires=acquires,
            assigns=assigns,
            returns=return_vars,
            shots=shots,
            calibration_id=self.model.calibration_id,
        )


class QbloxCFGWalker(DfsTraversal):
    def __init__(self, contexts: dict[PulseChannel, QbloxContext]):
        super().__init__()
        self.contexts = contexts

    def enter(self, block):
        iterator = block.iterator()
        while (inst := next(iterator, None)) is not None:
            if isinstance(inst, Sweep):
                QbloxContext.enter_sweep(inst, self.contexts)
            elif isinstance(inst, Repeat):
                QbloxContext.enter_repeat(inst, self.contexts)
            elif isinstance(inst, QuantumInstruction):
                if isinstance(inst, PostProcessing):
                    continue
                elif isinstance(inst, Synchronize):
                    QbloxContext.synchronize(inst, self.contexts)
                    continue

                for target in inst.quantum_targets:
                    context = self.contexts[target]
                    if isinstance(inst, DeviceUpdate):
                        context.device_update(inst)
                    elif isinstance(inst, PhaseReset):
                        context.reset_phase()
                    elif isinstance(inst, MeasurePulse):
                        next_inst = next(iterator, None)
                        if next_inst is None or not isinstance(next_inst, Acquire):
                            raise ValueError(
                                "Found a MeasurePulse but no Acquire instruction followed"
                            )

                        context.measure_acquire(inst, next_inst, target)
                    elif isinstance(inst, Waveform):
                        context.waveform(inst, target)
                    elif isinstance(inst, Delay):
                        context.delay(inst)
                    elif isinstance(inst, PhaseShift):
                        context.shift_phase(inst)
                    elif isinstance(inst, FrequencyShift):
                        context.shift_frequency(inst, target)
                    elif isinstance(inst, Id):
                        context.id()

    def exit(self, block):
        iterator = block.iterator()
        while (inst := next(iterator, None)) is not None:
            if isinstance(inst, Repeat):
                QbloxContext.exit_repeat(inst, self.contexts)
            elif isinstance(inst, Sweep):
                QbloxContext.exit_sweep(inst, self.contexts)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd

from abc import ABC
from bisect import insort
from collections import OrderedDict, defaultdict
from contextlib import ExitStack, contextmanager
from dataclasses import dataclass, field
from itertools import chain
from numbers import Number
from typing import Dict, List, Optional, Tuple, Union

import numpy as np

from qat.purr.backends.qblox.analysis_passes import (
    BindingPass,
    BindingResult,
    CFGPass,
    CFGResult,
    IterBound,
    QbloxLegalisationPass,
    ReadWriteResult,
    ScopingResult,
    TILegalisationPass,
    TriagePass,
    TriageResult,
)
from qat.purr.backends.qblox.codegen_base import DfsTraversal
from qat.purr.backends.qblox.config import ModuleConfig, SequencerConfig
from qat.purr.backends.qblox.constants import Constants
from qat.purr.backends.qblox.ir import Opcode, Sequence, SequenceBuilder
from qat.purr.backends.qblox.transform_passes import ScopePeeling
from qat.purr.backends.utilities import evaluate_shape
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.devices import PulseChannel, PulseShapeType
from qat.purr.compiler.execution import DeviceInjectors, SweepIterator
from qat.purr.compiler.instructions import (
    Acquire,
    CustomPulse,
    Delay,
    DeviceUpdate,
    FrequencyShift,
    Id,
    Instruction,
    MeasurePulse,
    PhaseReset,
    PhaseShift,
    PostProcessing,
    Pulse,
    QuantumInstruction,
    Repeat,
    Sweep,
    Synchronize,
    Variable,
    Waveform,
    calculate_duration,
)
from qat.purr.core.metrics_base import MetricsManager
from qat.purr.core.pass_base import AnalysisPass, InvokerMixin, PassManager
from qat.purr.core.result_base import ResultInfoMixin, ResultManager
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


@dataclass
class QbloxPackage:
    pulse_channel_id: Optional[str] = None
    physical_channel_id: Optional[str] = None
    instrument_id: Optional[str] = None
    seq_idx: Optional[int] = None
    seq_config: SequencerConfig = field(default_factory=lambda: SequencerConfig())
    slot_idx: Optional[int] = None
    mod_config: ModuleConfig = field(default_factory=lambda: ModuleConfig())
    sequence: Optional[Sequence] = None
    timeline: Optional[np.ndarray] = None


@dataclass
class AllocationManager:
    _reg_pool: List[str] = field(
        default_factory=lambda: sorted(
            f"R{index}" for index in range(Constants.NUMBER_OF_REGISTERS)
        )
    )
    _lbl_counters: Dict[str, int] = field(default_factory=dict)

    registers: Dict[str, str] = field(default_factory=dict)
    labels: Dict[str, str] = field(default_factory=dict)

    def reg_alloc(self, name: str) -> str:
        if name in self.registers:
            log.warning(f"Returning a register already allocated for Variable {name}")
            register = self.registers[name]
        elif len(self._reg_pool) < 1:
            raise IndexError(
                "Out of registers. Attempting to use more registers than available in the Q1 sequence processor"
            )
        else:
            register = self._reg_pool.pop(0)
            self.registers[name] = register

        return register

    def reg_free(self, register: str) -> None:
        if register in self._reg_pool:
            raise RuntimeError(f"Cannot free register '{register}' as it's not in use")
        insort(self._reg_pool, register)
        self.registers = {
            var: reg for var, reg in self.registers.items() if reg != register
        }

    @contextmanager
    def reg_borrow(self, name: str):
        """
        Short-lived register allocation
        """

        register = self.reg_alloc(name)
        yield register
        self.reg_free(register)

    def label_gen(self, name: str):
        counter = self._lbl_counters.setdefault(name, 0)
        self._lbl_counters[name] += 1
        label = f"block_{name}_{counter}"
        self.labels[name] = label
        return label


class BaseEmitter:
    def __init__(self):
        self.allocations: Dict[int, Dict[str, int]] = defaultdict(dict)

    def allocate(self, target: PulseChannel) -> Tuple[int, int]:
        slot_idx = target.physical_channel.slot_idx
        target2seq = self.allocations[slot_idx]
        if (seq_idx := target2seq.get(target.full_id(), None)) is None:
            total = set(target.physical_channel.config.sequencers.keys())
            available = total - set(target2seq.values())
            if not available:
                raise ValueError(f"""
                No more available sequencers for {target}
                Physical channel: {target.physical_channel}
                Total available: {total}
                Current allocations: {target2seq}
                """)
            seq_idx = next(iter(available))
            target2seq[target.full_id()] = seq_idx
        return seq_idx, slot_idx


class QbloxContext(ABC):
    def __init__(
        self,
        rw_result: ReadWriteResult,
        iter_bounds: Dict[str, IterBound],
        alloc_mgr: AllocationManager = None,
    ):
        self.inits: Dict[str, List[Instruction]] = rw_result.inits
        self.reads: Dict[str, List[Instruction]] = rw_result.reads
        self.writes: Dict[str, List[Instruction]] = rw_result.writes
        self.iter_bounds = iter_bounds
        self.alloc_mgr = alloc_mgr or AllocationManager()

        self.sequence_builder = SequenceBuilder()
        self.sequencer_config = SequencerConfig()

        self._num_hw_avg = 1  # Technically disabled
        self._wf_memory: int = Constants.MAX_SAMPLE_SIZE_WAVEFORMS
        self._wf_index: int = 0
        self._acq_index: int = 0
        self._wgt_index: int = 0

        self._durations: List[Union[float, str]] = []
        self._frequency: float = 0.0  # Tracks frequency shifts on the target
        self._phases: List[Union[float, str]] = []
        self._timeline: np.ndarray = np.empty(0, dtype=complex)

    def __enter__(self):
        """
        Serves as the prologue
        """

        self.clear()

        self.sequence_builder.set_mrk(3)
        self.sequence_builder.set_latch_en(1, 4)
        self.sequence_builder.upd_param(Constants.GRID_TIME)

        for name, register in self.alloc_mgr.registers.items():
            self.sequence_builder.move(
                0,
                register,
                f"Precautionary initialisation for variable {name}",
            )

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Serves as the epilogue
        """

        self.sequence_builder.stop()

    @property
    def durations(self):
        return self._durations

    @property
    def duration(self):
        if all((isinstance(d, Number) for d in self._durations)):
            return sum(self._durations)

        ValueError("Cannot determine duration statically in dynamic settings")

    @property
    def phase(self):
        if all((isinstance(p, Number) for p in self._phases)):
            return sum(self._phases)

        ValueError("Cannot determine phase statically in dynamic settings")

    def create_package(
        self, target: PulseChannel, seq_idx: int, slot_idx: int
    ) -> QbloxPackage:
        lo_freq, nco_freq = TILegalisationPass.decompose_freq(target.frequency, target)

        qblox_config = target.physical_channel.config

        # Customise Module config
        mod_config: ModuleConfig = qblox_config.module
        if mod_config.lo.out0_en:
            mod_config.lo.out0_freq = lo_freq
        if mod_config.lo.out1_en:
            mod_config.lo.out1_freq = lo_freq
        if mod_config.lo.out0_in0_en:
            mod_config.lo.out0_in0_freq = lo_freq

        mod_config.scope_acq.sequencer_select = seq_idx
        mod_config.scope_acq.trigger_mode_path0 = "sequencer"
        mod_config.scope_acq.avg_mode_en_path0 = True
        mod_config.scope_acq.trigger_mode_path1 = "sequencer"
        mod_config.scope_acq.avg_mode_en_path1 = True

        # Customise Sequencer config
        seq_config: SequencerConfig = qblox_config.sequencers[seq_idx]
        seq_config.nco.freq = nco_freq
        seq_config.nco.prop_delay_comp_en = True
        seq_config.square_weight_acq.integration_length = (
            self.sequencer_config.square_weight_acq.integration_length
        )
        seq_config.thresholded_acq.rotation = self.sequencer_config.thresholded_acq.rotation
        seq_config.thresholded_acq.threshold = (
            self.sequencer_config.thresholded_acq.threshold
        )

        return QbloxPackage(
            pulse_channel_id=target.full_id(),
            physical_channel_id=target.physical_channel_id,
            seq_idx=seq_idx,
            seq_config=seq_config,
            slot_idx=slot_idx,
            mod_config=mod_config,
            sequence=self.sequence_builder.build(),
            timeline=self._timeline,
        )

    def is_empty(self):
        """
        Masks away yet-to-be-supported second-state, cancellation, and cross cancellation targets
        This is temporary and criteria will change with more features coming in
        """
        return not (
            self.sequence_builder.waveforms
            or self.sequence_builder.acquisitions
            or [
                inst
                for inst in self.sequence_builder.q1asm_instructions
                if inst.opcode
                in [
                    Opcode.PLAY,
                    Opcode.SET_AWG_OFFSET,
                    Opcode.SET_AWG_GAIN,
                    Opcode.ACQUIRE,
                    Opcode.ACQUIRE_TTL,
                    Opcode.ACQUIRE_WEIGHED,
                ]
            ]
        )

    def clear(self):
        self.sequence_builder.waveforms.clear()
        self.sequence_builder.acquisitions.clear()
        self.sequence_builder.weights.clear()
        self.sequence_builder.q1asm_instructions.clear()

    def wait_imm(self, duration: int):
        """
        Waits for `duration` nanoseconds expressed as an immediate.

        `duration` must be positive
        """

        if duration <= 0:
            log.debug(f"Expected positive duration, but got {duration}")
            return

        quotient = duration // Constants.MAX_WAIT_TIME
        remainder = duration % Constants.MAX_WAIT_TIME

        if remainder < Constants.GRID_TIME:
            log.debug(f"Rounding up {remainder} ns to {Constants.GRID_TIME} ns")
            remainder = Constants.GRID_TIME
            duration = quotient * Constants.MAX_WAIT_TIME + remainder

        if quotient >= Constants.LOOP_UNROLL_THRESHOLD:
            with self._loop("wait_label", quotient):
                self.sequence_builder.wait(Constants.MAX_WAIT_TIME)
        elif quotient >= 1:
            for _ in range(quotient):
                self.sequence_builder.wait(Constants.MAX_WAIT_TIME)

        self.sequence_builder.wait(remainder)

    def wait_reg(self, duration: str):
        """
        A mere `wait RX` in general has undefined runtime behaviour.

        This is a useful helper to dynamically wait for a `duration` nanoseconds expressed as a register.
        `iter_reg` is a short-lived register only used as an interator. Customer is responsible for (de)allocation.
        """

        batch_enter = self.alloc_mgr.label_gen("batch_enter")
        batch_exit = self.alloc_mgr.label_gen("batch_exit")
        remainder = self.alloc_mgr.label_gen("remainder")
        round_up = self.alloc_mgr.label_gen("round_up")

        with self.alloc_mgr.reg_borrow("tmp") as iter_reg:
            if duration != iter_reg:
                self.sequence_builder.move(
                    duration,
                    iter_reg,
                    f"Avoids cluttering {duration} as it's likely used as an accumulator",
                )
                self.sequence_builder.nop()

            self.sequence_builder.jlt(iter_reg, Constants.GRID_TIME, round_up)
            self.sequence_builder.jlt(iter_reg, Constants.MAX_WAIT_TIME, remainder)
            self.sequence_builder.label(batch_enter)
            self.sequence_builder.wait(Constants.MAX_WAIT_TIME)
            self.sequence_builder.sub(iter_reg, Constants.MAX_WAIT_TIME, iter_reg)
            self.sequence_builder.nop()
            self.sequence_builder.jge(iter_reg, Constants.MAX_WAIT_TIME, batch_enter)
            self.sequence_builder.jge(iter_reg, Constants.GRID_TIME, remainder)
            self.sequence_builder.label(round_up)
            self.sequence_builder.jlt(iter_reg, 1, batch_exit)
            self.sequence_builder.move(Constants.GRID_TIME, iter_reg, "Rounding up")
            self.sequence_builder.nop()
            self.sequence_builder.label(remainder)
            self.sequence_builder.wait(iter_reg)
            self.sequence_builder.label(batch_exit)

    def ledger(self, duration: Union[int, str], pulse: np.ndarray = None):
        """
        A helper method to keep track of the durations and phase as the codegen runs.

        Every little bit of time is accounted for and a timeline object (in static cases)
        is also kept up-to-date for debugging and visualisation purposes.
        """

        if isinstance(duration, int):
            pulse = np.zeros(duration, dtype=complex) if pulse is None else pulse
            if all((isinstance(p, Number) for p in self._phases)):
                total_phase = sum(self._phases)
                pulse = pulse * np.exp(1.0j * total_phase)
            assert pulse.size == duration
            self._durations.append(duration)
            if self._timeline is not None:
                self._timeline = np.append(self._timeline, pulse)
        elif isinstance(duration, str):
            self._durations.append(duration)
            self._timeline = None  # Destroy the timeline as we are in dynamic setting
        else:
            raise ValueError(f"Expected legal immediate or register, but got {duration}")

    @contextmanager
    def _loop(self, label: str, count: int = 1):
        """
        This context can be used to avoid unrolling loops but needs manual addition of the 0 iteration
        either immediately after the yield or immediately after the context has exited.
        """

        label = self.alloc_mgr.label_gen(label)
        register = self.alloc_mgr.reg_alloc(label)
        self.sequence_builder.move(count, register)
        self.sequence_builder.nop()
        self.sequence_builder.label(label)
        yield register
        self.sequence_builder.loop(register, label)
        self.alloc_mgr.reg_free(register)

    @contextmanager
    def _modulo_reg(self, a: str, n: int):
        """
        Helpful context that prepares a temporary register b such that a ≅ b [n] where
        n is some immediate constant batch size.
        """

        if not isinstance(n, int) or n < 2:
            raise ValueError(
                f"Invalid batch size. Expected an integer >= 2, but got {n} instead"
            )

        batch_enter = self.alloc_mgr.label_gen("batch_enter")
        batch_exit = self.alloc_mgr.label_gen("batch_exit")

        iter_reg = self.alloc_mgr.reg_alloc("tmp")
        if a != iter_reg:
            self.sequence_builder.move(
                a,
                iter_reg,
                f"Avoids cluttering {a} as it's likely used as an accumulator",
            )
            self.sequence_builder.nop()

        self.sequence_builder.jlt(iter_reg, n, batch_exit)
        self.sequence_builder.label(batch_enter)
        self.sequence_builder.sub(iter_reg, n, iter_reg)
        self.sequence_builder.nop()
        self.sequence_builder.jge(iter_reg, n, batch_enter)
        self.sequence_builder.label(batch_exit)
        yield iter_reg
        self.alloc_mgr.reg_free(iter_reg)

    def _upd_param_imm(self, duration: int):
        self.sequence_builder.upd_param(Constants.GRID_TIME)
        self.wait_imm(duration - Constants.GRID_TIME)

    def _upd_param_reg(self, duration: str):
        """
        Update latched parameters and wait for `duration` nanoseconds expressed as a register

        If `duration` is less than the threshold Constants.GRID_TIME, the generates assembly rounds up
        and waits for Constants.GRID_TIME.

        Note that the `up_param` instruction is defined only for an immediate operand which must be
        at least Constants.GRID_TIME. Therefore, technically we're already forced to round up.
        """

        batch_enter = self.alloc_mgr.label_gen("batch_enter")
        batch_exit = self.alloc_mgr.label_gen("batch_exit")
        remainder = self.alloc_mgr.label_gen("remainder")

        self.sequence_builder.upd_param(Constants.GRID_TIME)
        with self.alloc_mgr.reg_borrow("tmp") as iter_reg:
            self.sequence_builder.jlt(
                duration,
                Constants.GRID_TIME,
                batch_exit,
                "Guards against risky underflow",
            )
            self.sequence_builder.sub(duration, Constants.GRID_TIME, iter_reg)
            self.sequence_builder.nop()

            self.sequence_builder.jlt(iter_reg, Constants.GRID_TIME, batch_exit)
            self.sequence_builder.jlt(iter_reg, Constants.MAX_WAIT_TIME, remainder)
            self.sequence_builder.label(batch_enter)
            self.sequence_builder.wait(Constants.MAX_WAIT_TIME)
            self.sequence_builder.sub(iter_reg, Constants.MAX_WAIT_TIME, iter_reg)
            self.sequence_builder.nop()
            self.sequence_builder.jge(iter_reg, Constants.MAX_WAIT_TIME, batch_enter)
            self.sequence_builder.jlt(iter_reg, Constants.GRID_TIME, batch_exit)
            self.sequence_builder.label(remainder)
            self.sequence_builder.wait(iter_reg)
            self.sequence_builder.label(batch_exit)

    def _evaluate_waveform(self, waveform: Waveform, target: PulseChannel):
        """
        The waveform is evaluated as a 1d complex array. In QBlox, the Real and Imag parts of the pulse
        represent the digital offset on the AWG. They both must be within range [-1, 1].
        """

        num_samples = int(calculate_duration(waveform))
        dt = target.sample_time
        length = num_samples * dt
        centre = length / 2.0
        t = np.linspace(
            start=-centre + 0.5 * dt, stop=length - centre - 0.5 * dt, num=num_samples
        )
        pulse = evaluate_shape(waveform, t)
        scale = target.scale
        if isinstance(waveform, (Pulse, CustomPulse)) and waveform.ignore_channel_scale:
            scale = 1

        pulse *= scale
        pulse += target.bias

        if ((pulse.real < -1) | (pulse.real > 1)).any():
            raise ValueError(
                "Voltage range for I exceeded. Make sure all I values are within the range [-1, 1]"
            )
        if ((pulse.imag < -1) | (pulse.imag > 1)).any():
            raise ValueError(
                "Voltage range for Q exceeded. Make sure all I values are within the range [-1, 1]"
            )

        return pulse

    def _evaluate_square_pulse(self, pulse: Pulse, target: PulseChannel):
        bias = target.bias
        scale = 1.0 + 0.0j if pulse.ignore_channel_scale else target.scale

        pulse_amp = pulse.scale_factor * pulse.amp
        pulse_amp = scale * pulse_amp + bias

        if np.abs(pulse_amp.real) > 1 or np.abs(pulse_amp.imag) > 1:
            raise ValueError(
                f"""
                Voltage range for I or Q exceeded. Current values:
                waveform.scale_factor: {pulse.scale_factor}
                waveform.amp: {pulse.amp}
                (Channel) scale: {scale}
                (Channel) bias: {bias}
                """
            )

        return pulse_amp

    def _register_signal(self, waveform, target, data, name):
        index = self.sequence_builder.lookup_waveform_by_data(data)
        if index is not None:
            log.debug(
                f"Reusing signal {name} at index {index} for pulse {waveform} on channel {target}"
            )
        elif data.size > self._wf_memory:
            raise ValueError(
                f"No more waveform memory left for signal {name} of pulse {waveform} on channel {target}"
            )
        else:
            wf_hash = hash(waveform)
            wf_name = type(waveform).__name__
            if isinstance(waveform, Pulse):
                wf_name = waveform.shape.name

            index = self._wf_index
            self.sequence_builder.add_waveform(
                f"{wf_name}_{wf_hash}_{name}", index, data.tolist()
            )
            self._wf_memory = self._wf_memory - data.size
            self._wf_index = self._wf_index + 1

        return index

    def _register_waveform(self, waveform, target, data):
        i_index = self._register_signal(waveform, target, data.real, "I")
        q_index = self._register_signal(waveform, target, data.imag, "Q")
        return i_index, q_index

    def _register_acquisition(self, acq_name: str, num_bins: Union[int, str]):
        acq_index = self._acq_index
        self.sequence_builder.add_acquisition(acq_name, acq_index, num_bins)
        self._acq_index = acq_index + 1
        return acq_index

    def _register_weight(self, name: str, weight: np.ndarray):
        wgt_index = self._wgt_index
        self.sequence_builder.add_weight(name, wgt_index, weight.tolist())
        self._wgt_index = wgt_index + 1
        return wgt_index

    def _do_acquire(self, wgt_name, acq_filter, acq_index, acq_bin, acq_width):
        legal_acq_width = min(acq_width, Constants.MAX_WAIT_TIME)
        if acq_filter:
            weight = acq_filter.samples
            wgt_i_index = self._register_weight(f"{wgt_name}_weight_I", weight.real)
            wgt_q_index = self._register_weight(f"{wgt_name}_weight_Q", weight.imag)

            # TODO - silly convoluted way of using the `acquire_weighed` instruction
            # TODO - simplify when Qblox supports more freedom of memory addressing
            with (
                self.alloc_mgr.reg_borrow("wgt_i") as wgt_i_reg,
                self.alloc_mgr.reg_borrow("wgt_q") as wgt_q_reg,
            ):
                self.sequence_builder.move(wgt_i_index, wgt_i_reg)
                self.sequence_builder.move(wgt_q_index, wgt_q_reg)

                self.sequence_builder.acquire_weighed(
                    acq_index, acq_bin, wgt_i_reg, wgt_q_reg, legal_acq_width
                )
        else:
            self.sequence_builder.acquire(acq_index, acq_bin, legal_acq_width)

        self.wait_imm(acq_width - legal_acq_width)

    @contextmanager
    def _wrapper_pulse(
        self,
        delay_width,
        pulse_width,
        pulse_shape,
        i_steps,
        q_steps,
        i_index,
        q_index,
    ):
        effective_width = max(min(pulse_width, delay_width), Constants.GRID_TIME)
        if pulse_shape == PulseShapeType.SQUARE:
            self.sequence_builder.set_awg_offs(i_steps, q_steps)
            self.sequence_builder.upd_param(effective_width)
        else:
            self.sequence_builder.play(i_index, q_index, effective_width)
        remaining_width = pulse_width - effective_width
        if remaining_width < Constants.GRID_TIME:
            raise ValueError(
                f"""
                Remaining width after delay must be at least {Constants.GRID_TIME} ns.
                Got pulse width = {pulse_width} ns, delay width = {delay_width} ns. This leaves {remaining_width} ns,
                which isn't enough to trigger acquisition afterwards. Adjust the pulse width or delay width values so
                that the following holds:
                pulse_width - max(min(pulse_width, delay_width), Constants.GRID_TIME) >= Constants.GRID_TIME
                """
            )
        yield remaining_width
        if pulse_shape == PulseShapeType.SQUARE:
            self.sequence_builder.set_awg_offs(0, 0)
            self.sequence_builder.upd_param(Constants.GRID_TIME)

            self.ledger(Constants.GRID_TIME)

    @contextmanager
    def _wrapper_cond(self, mask, operator, duration):
        """
        A wrapper for conditional regions
        """

        self.sequence_builder.set_cond(
            1, mask, operator, duration, "Start of conditional region"
        )
        yield
        self.sequence_builder.set_cond(
            0, mask, operator, duration, "End of conditional region"
        )

    def id(self):
        self.sequence_builder.nop()

    def delay(self, inst: Delay):
        if isinstance(inst.duration, Variable):
            duration = self.alloc_mgr.registers[inst.duration.name]
            self.wait_reg(duration)
        elif isinstance(inst.duration, Number):
            duration = int(calculate_duration(inst))
            self.wait_imm(duration)
        else:
            raise ValueError(f"Expected a Variable or a Number but got {inst.duration}")

        self.ledger(duration)

    def shift_frequency(self, inst, target):
        # TODO - discuss w/t quantum_target.fixed_if is True or False
        old_frequency = target.frequency + self._frequency
        new_frequency = old_frequency + inst.frequency
        if new_frequency < target.min_frequency or new_frequency > target.max_frequency:
            raise ValueError(
                f"Cannot shift pulse channel frequency from '{old_frequency}' to '{new_frequency}'"
            )

        self._frequency = self._frequency + inst.frequency
        shifted_nco_freq = target.baseband_if_frequency + self._frequency
        value = QbloxLegalisationPass.freq_as_steps(shifted_nco_freq)  # 1 MHZ <-> 4e6
        self.sequence_builder.set_freq(value)
        self.sequence_builder.upd_param(Constants.GRID_TIME)

        self.ledger(Constants.GRID_TIME)

    def shift_phase(self, inst: PhaseShift):
        if isinstance(inst.phase, Variable):
            ph_reg = self.alloc_mgr.registers[inst.phase.name]
            with self._modulo_reg(ph_reg, Constants.NCO_MAX_PHASE_STEPS) as iter_reg:
                self.sequence_builder.set_ph_delta(iter_reg)
                self.sequence_builder.upd_param(Constants.GRID_TIME)
        elif isinstance(inst.phase, Number):
            ph_imm = QbloxLegalisationPass.phase_as_steps(inst.phase)
            self.sequence_builder.set_ph_delta(ph_imm)
            self.sequence_builder.upd_param(Constants.GRID_TIME)
        else:
            raise ValueError(f"Expected a Variable or a Number but got {inst.phase}")

        self._phases.append(inst.phase)
        self.ledger(Constants.GRID_TIME)

    def reset_phase(self):
        self.sequence_builder.reset_ph()
        self.sequence_builder.upd_param(Constants.GRID_TIME)

        self._phases.clear()
        self.ledger(Constants.GRID_TIME)

    def waveform(self, waveform: Waveform, target: PulseChannel):
        attr2var = {
            attr: var for attr, var in vars(waveform).items() if isinstance(var, Variable)
        }

        if isinstance(waveform, Pulse) and waveform.shape == PulseShapeType.SQUARE:
            if "width" in attr2var:
                pulse_width = attr2var["width"]
                pulse_width = self.alloc_mgr.registers[pulse_width.name]
                # TODO - Continue semantics: jump back to beginning of loop
            else:
                pulse_width = int(calculate_duration(waveform))
                if pulse_width < Constants.GRID_TIME:
                    log.debug(
                        f"""
                        Minimum pulse width is {Constants.GRID_TIME} ns with a resolution of {1} ns.
                        Please round up the width to at least {Constants.GRID_TIME} nanoseconds.
                        This pulse will be ignored.
                        """
                    )
                    return

            if "amp" in attr2var:
                pulse_amp = attr2var["amp"]
                i_steps = self.alloc_mgr.registers[pulse_amp.name]
                q_steps = self.alloc_mgr.registers["zero"]
            else:
                pulse_amp = self._evaluate_square_pulse(waveform, target)
                i_steps = int(pulse_amp.real * Constants.MAX_OFFSET)
                q_steps = int(pulse_amp.imag * Constants.MAX_OFFSET)

            self.sequence_builder.set_awg_offs(i_steps, q_steps)
            if "width" in attr2var:
                self._upd_param_reg(pulse_width)
            else:
                self._upd_param_imm(pulse_width)
            self.sequence_builder.set_awg_offs(0, 0)
            self.sequence_builder.upd_param(Constants.GRID_TIME)

            self.ledger(Constants.GRID_TIME)

            pulse = (
                np.ones(pulse_width, dtype=complex) * pulse_amp
                if isinstance(pulse_width, Number) and isinstance(pulse_amp, Number)
                else None
            )
        else:
            pulse = self._evaluate_waveform(waveform, target)
            pulse_width = pulse.size
            if pulse_width < Constants.GRID_TIME:
                log.debug(
                    f"""
                    Minimum pulse width is {Constants.GRID_TIME} ns with a resolution of {1} ns.
                    Please round up the width to at least {Constants.GRID_TIME} nanoseconds.
                    This pulse will be ignored.
                    """
                )
                return

            max_width = min(pulse_width, Constants.MAX_WAIT_TIME)
            i_index, q_index = self._register_waveform(waveform, target, pulse)
            self.sequence_builder.play(i_index, q_index, max_width)
            self.wait_imm(pulse_width - max_width)

        self.ledger(pulse_width, pulse)

    def measure_acquire(
        self, measure: MeasurePulse, acquire: Acquire, target: PulseChannel
    ):
        """
        Regarding HW thresholding, `acquire.rotation` is in radians and `acquire.threshold` is uncorrected.
        (Exactly as they've been computed during measure block construction)

        Here at configuration time, we "legalise" these parameters to what Qblox requires:
            + `rotation` as degrees
            + `threshold` to be corrected by the integration length
        """

        pulse = self._evaluate_waveform(measure, target)
        pulse_width = pulse.size
        delay_width = int(calculate_duration(Delay(acquire.channel, acquire.delay)))

        if pulse_width < Constants.GRID_TIME:
            log.debug(
                f"""
                Minimum pulse width is {Constants.GRID_TIME} ns with a resolution of {1} ns.
                Please round up the width to at least {Constants.GRID_TIME} nanoseconds.
                This pulse will be ignored. 
                """
            )
            return

        if pulse_width < delay_width + Constants.GRID_TIME:
            raise ValueError(
                f"""
                Expected pulse width >= delay width + {Constants.GRID_TIME} ns.
                Got pulse width = {pulse_width} ns and delay width = {delay_width} ns.
                """
            )

        name = f"{hash(acquire)}"
        num_bins = self.iter_bounds[name].count
        acq_bin = self.alloc_mgr.registers[name]
        acq_width = int(calculate_duration(acquire))
        acq_index = self._register_acquisition(acquire.output_variable, num_bins)
        self.sequencer_config.square_weight_acq.integration_length = acq_width
        self.sequencer_config.thresholded_acq.rotation = np.rad2deg(acquire.rotation)
        self.sequencer_config.thresholded_acq.threshold = acq_width * acquire.threshold

        i_steps, q_steps = None, None
        i_index, q_index = None, None
        if measure.shape == PulseShapeType.SQUARE:
            i_steps = int(pulse[0].real * Constants.MAX_OFFSET)
            q_steps = int(pulse[0].imag * Constants.MAX_OFFSET)
        else:
            i_index, q_index = self._register_waveform(measure, target, pulse)

        with self._wrapper_pulse(
            delay_width, pulse_width, measure.shape, i_steps, q_steps, i_index, q_index
        ) as remaining_width:
            self._do_acquire(
                acquire.output_variable,
                acquire.filter,
                acq_index,
                acq_bin,
                remaining_width,
            )

        self.ledger(pulse_width, pulse)

    @staticmethod
    def synchronize(inst: Synchronize, contexts: Dict):
        """
        Favours static time padding whenever possible, or else uses SYNC.
        TODO - Default to SYNC when Qblox supports finer grained SYNC groups
        """

        is_static = all(
            all((isinstance(d, Number) for d in contexts[target].durations))
            for target in inst.quantum_targets
        )

        for target in inst.quantum_targets:
            cxt = contexts[target]
            if is_static:
                duration_offset = (
                    max((contexts[target].duration for target in inst.quantum_targets))
                    - cxt.duration
                )
                cxt.wait_imm(duration_offset)
                cxt.ledger(duration_offset)
            else:
                cxt.sequence_builder.wait_sync(Constants.GRID_TIME)
                cxt.ledger(Constants.GRID_TIME)

    def device_update(self, du_inst: DeviceUpdate):
        if isinstance(du_inst.value, Variable):
            val_reg = self.alloc_mgr.registers[du_inst.value.name]
            if du_inst.attribute == "frequency":
                self.sequence_builder.set_freq(val_reg)
                self.sequence_builder.upd_param(Constants.GRID_TIME)

                self.ledger(Constants.GRID_TIME)
            else:
                raise NotImplementedError(
                    f"Unsupported processing of attribute {du_inst.attribute} for instruction {du_inst}"
                )
        elif isinstance(du_inst.value, Number):
            if du_inst.attribute == "frequency":
                lo_freq, nco_freq = TILegalisationPass.decompose_freq(
                    du_inst.value, du_inst.target
                )
                freq_imm = QbloxLegalisationPass.freq_as_steps(nco_freq)
                self.sequence_builder.set_freq(freq_imm)
                self.sequence_builder.upd_param(Constants.GRID_TIME)

                self.ledger(Constants.GRID_TIME)
            elif du_inst.attribute == "scale":
                pass
            else:
                raise NotImplementedError(
                    f"Unsupported processing of attribute {du_inst.attribute} for instruction {du_inst}"
                )
        else:
            raise ValueError(f"Expected a Variable or a Number but got {du_inst.value}")

    @staticmethod
    def enter_repeat(inst: Repeat, contexts: Dict):
        iter_name = f"{hash(inst)}"
        for context in contexts.values():
            var_names = set([n for (n, insts) in context.inits.items() if inst in insts])
            for name in var_names:
                register = context.alloc_mgr.registers[name]
                bound = context.iter_bounds[name]
                context.sequence_builder.move(
                    bound.start,
                    register,
                    f"Initialise register allocated for variable {name}",
                )

            label = context.alloc_mgr.labels[iter_name]
            context.sequence_builder.label(label)

            context.sequence_builder.wait_sync(
                Constants.GRID_TIME, "Sync at the beginning of shot"
            )
            context.reset_phase()

    @staticmethod
    def exit_repeat(inst: Repeat, contexts: Dict):
        iter_name = f"{hash(inst)}"
        for context in contexts.values():
            context.wait_imm(int(inst.repetition_period * 1e9))

            var_names = set(
                [n for (n, insts) in context.writes.items() if inst in insts]
                + [n for (n, insts) in context.reads.items() if inst in insts]
            )
            for name in var_names:
                register = context.alloc_mgr.registers[name]
                bound = context.iter_bounds[name]
                context.sequence_builder.add(
                    register,
                    bound.step,
                    register,
                    f"Increment register for variable {name}",
                )
                context.sequence_builder.nop()

            register = context.alloc_mgr.registers[iter_name]
            bound = context.iter_bounds[iter_name]
            label = context.alloc_mgr.labels[iter_name]
            context.sequence_builder.jlt(register, bound.end + bound.step, label)

    @staticmethod
    def enter_sweep(inst: Sweep, contexts: Dict):
        iter_name = f"{hash(inst)}"
        for context in contexts.values():
            var_names = set([n for (n, insts) in context.inits.items() if inst in insts])
            for name in var_names:
                register = context.alloc_mgr.registers[name]
                bound = context.iter_bounds[name]
                context.sequence_builder.move(
                    bound.start,
                    register,
                    f"Initialise register allocated for variable {name}",
                )

            label = context.alloc_mgr.labels[iter_name]
            context.sequence_builder.label(label)

    @staticmethod
    def exit_sweep(inst: Sweep, contexts: Dict):
        iter_name = f"{hash(inst)}"
        for context in contexts.values():
            var_names = set(
                [n for (n, insts) in context.writes.items() if inst in insts]
                + [n for (n, insts) in context.reads.items() if inst in insts]
            )
            for name in var_names:
                register = context.alloc_mgr.registers[name]
                bound = context.iter_bounds[name]
                context.sequence_builder.add(
                    register,
                    bound.step,
                    register,
                    f"Increment register for variable {name}",
                )
                context.sequence_builder.nop()

            register = context.alloc_mgr.registers[iter_name]
            bound = context.iter_bounds[iter_name]
            label = context.alloc_mgr.labels[iter_name]
            context.sequence_builder.jlt(register, bound.end + bound.step, label)


@dataclass
class PreCodegenResult(ResultInfoMixin):
    alloc_mgrs: Dict[PulseChannel, AllocationManager] = field(
        default_factory=lambda: defaultdict(lambda: AllocationManager())
    )


class PreCodegenPass(AnalysisPass):
    """
    Precedes code generation. The context-based emitter needs registers pre-allocated for every
    live variable in program. This pass performs a naive register allocation through a manager object.
    """

    def run(self, ir: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs):
        triage_result = res_mgr.lookup_by_type(TriageResult)
        binding_result = res_mgr.lookup_by_type(BindingResult)
        result = PreCodegenResult()

        for target in triage_result.target_map:
            alloc_mgr = result.alloc_mgrs[target]
            iter_bound_result = binding_result.iter_bound_results[target]
            reads = binding_result.rw_results[target].reads
            writes = binding_result.rw_results[target].writes

            alloc_mgr.reg_alloc("zero")
            names = set(chain(*[iter_bound_result.keys(), reads.keys(), writes.keys()]))
            for name in names:
                alloc_mgr.reg_alloc(name)
                alloc_mgr.label_gen(name)

        res_mgr.add(result)


class QbloxEmitter(BaseEmitter, InvokerMixin):
    def build_pass_pipeline(self, *args, **kwargs):
        return PassManager() | TriagePass() | BindingPass() | PreCodegenPass()

    def emit_packages(
        self,
        ir: InstructionBuilder,
        res_mgr: ResultManager,
        met_mgr: MetricsManager,
        ignore_empty=True,
    ) -> Dict[int, List[QbloxPackage]]:
        triage_result = res_mgr.lookup_by_type(TriageResult)
        device_updates = triage_result.device_updates
        quantum_instructions = triage_result.quantum_instructions
        sweeps = triage_result.sweeps
        switerator = SweepIterator.from_sweeps(sweeps)
        dinjectors = DeviceInjectors(device_updates)

        binding_result = res_mgr.lookup_by_type(BindingResult)
        scoping_result = next(
            iter(binding_result.scoping_results.values()), ScopingResult()
        )
        scopes = [(s, e) for (s, e) in scoping_result.scope2symbols if s in sweeps]

        ScopePeeling().run(ir, res_mgr, met_mgr, scopes=scopes)
        # TODO - Pass infrastructural improvements [COMPILER-849]
        # res_mgr.mark_as_dirty(triage_result, binding_result)

        iter2packages: Dict[int, List[QbloxPackage]] = OrderedDict()
        try:
            dinjectors.inject()
            while not switerator.is_finished():
                switerator.do_sweep(quantum_instructions)
                inner_res_mgr = ResultManager()
                inner_met_mgr = MetricsManager()
                self.run_pass_pipeline(ir, inner_res_mgr, inner_met_mgr)

                inner_triage_result = inner_res_mgr.lookup_by_type(TriageResult)
                inner_binding_result = inner_res_mgr.lookup_by_type(BindingResult)
                inner_precodegen_result = inner_res_mgr.lookup_by_type(PreCodegenResult)
                iter2packages[switerator.accumulated_sweep_iteration] = self._do_emit(
                    inner_triage_result,
                    inner_binding_result,
                    inner_precodegen_result,
                    ignore_empty,
                )
        finally:
            switerator.revert(triage_result.quantum_instructions)
            dinjectors.revert()

        return iter2packages

    def _do_emit(
        self,
        triage_result: TriageResult,
        binding_result: BindingResult,
        precodegen_result: PreCodegenResult,
        ignore_empty=True,
    ) -> List[QbloxPackage]:
        rw_results: Dict[PulseChannel, ReadWriteResult] = binding_result.rw_results
        iter_bound_results: Dict[PulseChannel, Dict[str, IterBound]] = (
            binding_result.iter_bound_results
        )
        alloc_mgrs: Dict[PulseChannel, AllocationManager] = precodegen_result.alloc_mgrs

        with ExitStack() as stack:
            contexts = {
                t: stack.enter_context(
                    QbloxContext(
                        rw_result=rw_results[t],
                        iter_bounds=iter_bound_results[t],
                        alloc_mgr=alloc_mgrs[t],
                    )
                )
                for t in triage_result.target_map
            }

            for repeat in triage_result.repeats:
                QbloxContext.enter_repeat(repeat, contexts)

            iterator = iter(triage_result.quantum_instructions)
            while (inst := next(iterator, None)) is not None:
                if isinstance(inst, PostProcessing):
                    continue
                elif isinstance(inst, Synchronize):
                    QbloxContext.synchronize(inst, contexts)
                    continue

                for target in inst.quantum_targets:
                    context = contexts[target]
                    if isinstance(inst, DeviceUpdate):
                        context.device_update(inst)
                    elif isinstance(inst, PhaseReset):
                        context.reset_phase()
                    elif isinstance(inst, MeasurePulse):
                        next_inst = next(iterator, None)
                        if next_inst is None or not isinstance(next_inst, Acquire):
                            raise ValueError(
                                "Found a MeasurePulse but no Acquire instruction followed"
                            )

                        context.measure_acquire(inst, next_inst, target)
                    elif isinstance(inst, Waveform):
                        context.waveform(inst, target)
                    elif isinstance(inst, Delay):
                        context.delay(inst)
                    elif isinstance(inst, PhaseShift):
                        context.shift_phase(inst)
                    elif isinstance(inst, FrequencyShift):
                        context.shift_frequency(inst, target)
                    elif isinstance(inst, Id):
                        context.id()

            for repeat in triage_result.repeats:
                QbloxContext.exit_repeat(repeat, contexts)

        packages = []
        for target, context in contexts.items():
            if not context.is_empty() or not ignore_empty:
                seq_idx, slot_idx = self.allocate(target)
                package = context.create_package(target, seq_idx, slot_idx)
                packages.append(package)
        return packages


class NewQbloxEmitter(BaseEmitter, InvokerMixin):
    def build_pass_pipeline(self, *args, **kwargs):
        return PassManager() | PreCodegenPass() | CFGPass()

    def emit_packages(
        self,
        ir: InstructionBuilder,
        res_mgr: ResultManager,
        met_mgr: MetricsManager,
        ignore_empty=True,
    ) -> List[QbloxPackage]:
        self.run_pass_pipeline(ir, res_mgr, met_mgr)

        triage_result = res_mgr.lookup_by_type(TriageResult)
        binding_result = res_mgr.lookup_by_type(BindingResult)
        precodegen_result = res_mgr.lookup_by_type(PreCodegenResult)
        cfg_result = res_mgr.lookup_by_type(CFGResult)

        rw_results: Dict[PulseChannel, ReadWriteResult] = binding_result.rw_results
        iter_bound_results: Dict[PulseChannel, Dict[str, IterBound]] = (
            binding_result.iter_bound_results
        )
        alloc_mgrs: Dict[PulseChannel, AllocationManager] = precodegen_result.alloc_mgrs

        with ExitStack() as stack:
            contexts = {
                t: stack.enter_context(
                    QbloxContext(
                        rw_result=rw_results[t],
                        iter_bounds=iter_bound_results[t],
                        alloc_mgr=alloc_mgrs[t],
                    )
                )
                for t in triage_result.target_map
            }
            QbloxCFGWalker(contexts).run(cfg_result.cfg)

        packages = []
        for target, context in contexts.items():
            if not context.is_empty() or not ignore_empty:
                seq_idx, slot_idx = self.allocate(target)
                package = context.create_package(target, seq_idx, slot_idx)
                packages.append(package)
        return packages


class QbloxCFGWalker(DfsTraversal):
    def __init__(self, contexts: Dict[PulseChannel, QbloxContext]):
        super().__init__()
        self.contexts = contexts

    def enter(self, block):
        iterator = block.iterator()
        while (inst := next(iterator, None)) is not None:
            if isinstance(inst, Sweep):
                QbloxContext.enter_sweep(inst, self.contexts)
            elif isinstance(inst, Repeat):
                QbloxContext.enter_repeat(inst, self.contexts)
            elif isinstance(inst, QuantumInstruction):
                if isinstance(inst, PostProcessing):
                    continue
                elif isinstance(inst, Synchronize):
                    QbloxContext.synchronize(inst, self.contexts)
                    continue

                for target in inst.quantum_targets:
                    context = self.contexts[target]
                    if isinstance(inst, DeviceUpdate):
                        context.device_update(inst)
                    elif isinstance(inst, PhaseReset):
                        context.reset_phase()
                    elif isinstance(inst, MeasurePulse):
                        next_inst = next(iterator, None)
                        if next_inst is None or not isinstance(next_inst, Acquire):
                            raise ValueError(
                                "Found a MeasurePulse but no Acquire instruction followed"
                            )

                        context.measure_acquire(inst, next_inst, target)
                    elif isinstance(inst, Waveform):
                        context.waveform(inst, target)
                    elif isinstance(inst, Delay):
                        context.delay(inst)
                    elif isinstance(inst, PhaseShift):
                        context.shift_phase(inst)
                    elif isinstance(inst, FrequencyShift):
                        context.shift_frequency(inst, target)
                    elif isinstance(inst, Id):
                        context.id()

    def exit(self, block):
        iterator = block.iterator()
        while (inst := next(iterator, None)) is not None:
            if isinstance(inst, Repeat):
                QbloxContext.exit_repeat(inst, self.contexts)
            elif isinstance(inst, Sweep):
                QbloxContext.exit_sweep(inst, self.contexts)
